@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2005 Jeroen Frijters
+  Copyright (C) 2005, 2006 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -27,9 +27,7 @@ import java.lang.reflect.Field;
 
 public abstract class VMField
 {
-    // fieldCookie must be package accessible (actually "assembly") to allow map.xml
-    // implementation of LibraryVMInterfaceImpl.getWrapperFromField() to access it.
-    protected Object fieldCookie;
+    public Object fieldCookie;
     protected Class declaringClass;
     protected boolean isPublic;
     protected int modifiers;
@@ -46,21 +44,10 @@ public abstract class VMField
 
     public final int getModifiers()
     {
-        return modifiers & 0x00DF;
-    }
-
-    public final boolean isEnumConstant()
-    {
-        return (modifiers & 0x4000) != 0;
-    }
-
-    public final boolean isSynthetic()
-    {
-        return (modifiers & 0x1000) != 0;
+        return modifiers;
     }
 
     public abstract Field newField();
-    public abstract void checkAccess(Object o, Class caller) throws IllegalAccessException;
     public abstract String getName();
     public abstract Class getType();
     public abstract String getSignature();
