@@ -1,45 +1,32 @@
-/* PlainDatagramSocketImpl.java -- Default DatagramSocket implementation
-   Copyright (C) 1998, 1999, 2001 Free Software Foundation, Inc.
+/*
+  Copyright (C) 2002, 2003, 2004, 2005, 2006 Jeroen Frijters
 
-This file is part of GNU Classpath.
+  This software is provided 'as-is', without any express or implied
+  warranty.  In no event will the authors be held liable for any damages
+  arising from the use of this software.
 
-GNU Classpath is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
-any later version.
- 
-GNU Classpath is distributed in the hope that it will be useful, but
-WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with GNU Classpath; see the file COPYING.  If not, write to the
-Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-02111-1307 USA.
-
-Linking this library statically or dynamically with other modules is
-making a combined work based on this library.  Thus, the terms and
-conditions of the GNU General Public License cover the whole
-combination.
-
-As a special exception, the copyright holders of this library give you
-permission to link this library with independent modules to produce an
-executable, regardless of the license terms of these independent
-modules, and to copy and distribute the resulting executable under
-terms of your choice, provided that you also meet, for each linked
-independent module, the terms and conditions of the license of that
-module.  An independent module is a module which is not derived from
-or based on this library.  If you modify this library, you may extend
-this exception to your version of the library, but you are not
-obligated to do so.  If you do not wish to do so, delete this
-exception statement from your version. */
+  Permission is granted to anyone to use this software for any purpose,
+  including commercial applications, and to alter it and redistribute it
+  freely, subject to the following restrictions:
 
+  1. The origin of this software must not be misrepresented; you must not
+     claim that you wrote the original software. If you use this software
+     in a product, an acknowledgment in the product documentation would be
+     appreciated but is not required.
+  2. Altered source versions must be plainly marked as such, and must not be
+     misrepresented as being the original software.
+  3. This notice may not be removed or altered from any source distribution.
 
+  Jeroen Frijters
+  jeroen@frijters.net
+  
+*/
 package gnu.java.net;
 
 import java.io.IOException;
 import java.net.*;
+import java.util.*;
+import cli.System.Net.IPAddress;
 import cli.System.Net.IPEndPoint;
 import cli.System.Net.Sockets.SocketOptionName;
 import cli.System.Net.Sockets.SocketOptionLevel;
@@ -50,62 +37,34 @@ import cli.System.Net.Sockets.ProtocolType;
 import cli.System.Net.Sockets.AddressFamily;
 import ikvm.lang.CIL;
 
-/**
-* This is the default socket implementation for datagram sockets.
-* It makes native calls to C routines that implement BSD style
-* SOCK_DGRAM sockets in the AF_INET family.
-*
-* @version 0.1
-*
-* @author Aaron M. Renn (arenn@urbanophile.com)
-*/
 public class PlainDatagramSocketImpl extends DatagramSocketImpl
 {
-    /*
-     * Static Variables
-     */
+    // Winsock Error Codes
+    private static final int WSAEMSGSIZE   = 10040;
+    private static final int WSAECONNRESET = 10054;
 
-    /**
-     * This is the actual underlying socket
-     */
-    private cli.System.Net.Sockets.Socket socket = new cli.System.Net.Sockets.Socket(
-        AddressFamily.wrap(AddressFamily.InterNetwork), 
-        SocketType.wrap(SocketType.Dgram),
-        ProtocolType.wrap(ProtocolType.Udp));
+    private cli.System.Net.Sockets.Socket socket;
 
-    /*************************************************************************/
-
-    /*
-     * Constructors
-     */
-
-    /**
-     * Default do nothing constructor
-     */
-    public PlainDatagramSocketImpl()
+    public PlainDatagramSocketImpl() throws IOException
     {
+        try
+        {
+            if(false) throw new cli.System.Net.Sockets.SocketException();
+            socket = new cli.System.Net.Sockets.Socket(
+                AddressFamily.wrap(AddressFamily.InterNetwork), 
+                SocketType.wrap(SocketType.Dgram),
+                ProtocolType.wrap(ProtocolType.Udp));
+        }
+        catch(cli.System.Net.Sockets.SocketException x)
+        {
+            throw new SocketException(x.getMessage());
+        }
     }
 
-    /*************************************************************************/
-
-    /*
-     * Instance Methods
-     */
-
-    /**
-     * Creates a new datagram socket
-     *
-     * @exception SocketException If an error occurs
-     */
     protected void create() throws SocketException
     {
     }
 
-    /*************************************************************************/
-
-    /**
-     * Closes the socket
-     */
     protected void close()
     {
         if(socket != null)
@@ -115,16 +74,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Binds this socket to a particular port and interface
-     *
-     * @param port The port to bind to
-     * @param addr The address to bind to
-     *
-     * @exception SocketException If an error occurs
-     */
     protected void bind(int port, InetAddress addr) throws SocketException
     {
         try
@@ -144,16 +93,7 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Sends a packet of data to a remote host
-     *
-     * @param packet The packet to send
-     *
-     * @exception IOException If an error occurs
-     */
-    protected void send(DatagramPacket packet) throws IOException
+    public void send(DatagramPacket packet) throws IOException
     {
         try		
         {
@@ -180,73 +120,59 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * What does this method really do?
-     */
     protected int peek(InetAddress addr) throws IOException
     {
         throw new IOException("Not Implemented Yet");
     }
 
-    /*************************************************************************/
-
-    /**
-     * Receives a UDP packet from the network
-     *
-     * @param packet The packet to fill in with the data received
-     *
-     * @exception IOException IOException If an error occurs
-     */
-    protected void receive(DatagramPacket packet) throws IOException
+    public void receive(DatagramPacket packet) throws IOException
     {
         byte[] data = packet.getData();
         cli.System.Net.EndPoint[] remoteEP = new cli.System.Net.EndPoint[] 
             {
                 new cli.System.Net.IPEndPoint(0, 0)
             };
-        try
+        for(;;)
         {
-            if(false) throw new cli.System.Net.Sockets.SocketException();
-            if(false) throw new cli.System.ObjectDisposedException("");
-            int length = socket.ReceiveFrom(data, packet.getOffset(), getDatagramPacketBufferLength(packet),
-                            cli.System.Net.Sockets.SocketFlags.wrap(cli.System.Net.Sockets.SocketFlags.None), remoteEP);
-            setDatagramPacketLength(packet, length);
-        }
-        catch(cli.System.Net.Sockets.SocketException x)
-        {
-            // If the buffer size was too small for the packet, ReceiveFrom receives the part of the packet
-            // that fits in the buffer and then throws an exception, so we have to ignore the exception in this case.
-            if(x.get_ErrorCode() != 10040) //WSAEMSGSIZE
+            try
             {
+                if(false) throw new cli.System.Net.Sockets.SocketException();
+                if(false) throw new cli.System.ObjectDisposedException("");
+                int length = socket.ReceiveFrom(data, packet.getOffset(), getDatagramPacketBufferLength(packet),
+                                cli.System.Net.Sockets.SocketFlags.wrap(cli.System.Net.Sockets.SocketFlags.None), remoteEP);
+                setDatagramPacketLength(packet, length);
+                break;
+            }
+            catch(cli.System.Net.Sockets.SocketException x)
+            {
+                if(x.get_ErrorCode() == WSAECONNRESET)
+                {
+                    // A previous send failed (i.e. the remote host responded with a ICMP that the port is closed) and
+                    // the winsock stack helpfully lets us know this, but we don't care so we just retry the receive.
+                    continue;
+                }
+                if(x.get_ErrorCode() == WSAEMSGSIZE)
+                {
+                    // The buffer size was too small for the packet, ReceiveFrom receives the part of the packet
+                    // that fits in the buffer and then throws an exception, so we have to ignore the exception in this case.
+                    break;
+                }
                 throw PlainSocketImpl.convertSocketExceptionToIOException(x);
             }
+            catch(cli.System.ObjectDisposedException x1)
+            {
+                throw new SocketException("Socket is closed");
+            }
         }
-        catch(cli.System.ObjectDisposedException x1)
-        {
-            throw new SocketException("Socket is closed");
-        }
-        int remoteIP = (int)((cli.System.Net.IPEndPoint)remoteEP[0]).get_Address().get_Address();
-        byte[] ipv4 = new byte[] { (byte)remoteIP, (byte)(remoteIP >> 8), (byte)(remoteIP >> 16), (byte)(remoteIP >> 24) };
-        InetAddress remoteAddress = InetAddress.getByAddress(ipv4);
-        packet.setAddress(remoteAddress);
-        packet.setPort(((cli.System.Net.IPEndPoint)remoteEP[0]).get_Port());
+        IPEndPoint endpoint = (IPEndPoint)remoteEP[0];
+        packet.setAddress(PlainSocketImpl.getInetAddressFromIPEndPoint(endpoint));
+        packet.setPort(endpoint.get_Port());
     }
 
     // these methods live in map.xml, because we need to access the package accessible fields in DatagramPacket
     private static native void setDatagramPacketLength(DatagramPacket packet, int length);
     private static native int getDatagramPacketBufferLength(DatagramPacket packet);
 
-    /*************************************************************************/
-
-    /**
-     * Joins a multicast group
-     *
-     * @param addr The group to join
-     *
-     * @exception IOException If an error occurs
-     */
     protected void join(InetAddress addr) throws IOException
     {
         try
@@ -254,7 +180,7 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
             if(false) throw new cli.System.Net.Sockets.SocketException();
             if(false) throw new cli.System.ArgumentException();
             if(false) throw new cli.System.ObjectDisposedException("");
-            socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.AddMembership), new MulticastOption(new cli.System.Net.IPAddress(PlainSocketImpl.getAddressFromInetAddress(addr))));
+            socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.AddMembership), new MulticastOption(PlainSocketImpl.getAddressFromInetAddress(addr)));
         }
         catch(cli.System.Net.Sockets.SocketException x)
         {
@@ -270,15 +196,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Leaves a multicast group
-     *
-     * @param addr The group to leave
-     *
-     * @exception IOException If an error occurs
-     */
     protected void leave(InetAddress addr) throws IOException
     {
         try
@@ -286,7 +203,7 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
             if(false) throw new cli.System.Net.Sockets.SocketException();
             if(false) throw new cli.System.ArgumentException();
             if(false) throw new cli.System.ObjectDisposedException("");
-            socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.DropMembership), new MulticastOption(new cli.System.Net.IPAddress(PlainSocketImpl.getAddressFromInetAddress(addr))));
+            socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.DropMembership), new MulticastOption(PlainSocketImpl.getAddressFromInetAddress(addr)));
         }
         catch(cli.System.Net.Sockets.SocketException x)
         {
@@ -302,15 +219,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Gets the Time to Live value for the socket
-     *
-     * @return The TTL value
-     *
-     * @exception IOException If an error occurs
-     */
     protected byte getTTL() throws IOException
     {
         try
@@ -329,15 +237,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Sets the Time to Live value for the socket
-     *
-     * @param ttl The new TTL value
-     *
-     * @exception IOException If an error occurs
-     */
     protected void setTTL(byte ttl) throws IOException
     {
         try
@@ -356,15 +255,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Gets the Time to Live value for the socket
-     *
-     * @return The TTL value
-     *
-     * @exception IOException If an error occurs
-     */
     protected int getTimeToLive() throws IOException
     {
         try
@@ -383,15 +273,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Sets the Time to Live value for the socket
-     *
-     * @param ttl The new TTL value
-     *
-     * @exception IOException If an error occurs
-     */
     protected void setTimeToLive(int ttl) throws IOException
     {
         try
@@ -410,17 +291,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Retrieves the value of an option on the socket
-     *
-     * @param option_id The identifier of the option to retrieve
-     *
-     * @return The value of the option
-     *
-     * @exception SocketException If an error occurs
-     */
     public Object getOption(int option_id) throws SocketException
     {
         try
@@ -429,8 +299,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
             if(false) throw new cli.System.ObjectDisposedException("");
             switch(option_id)
             {
-                case SocketOptions.SO_REUSEADDR:
-                    return new Boolean(CIL.unbox_int(socket.GetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.ReuseAddress))) != 0);
                 case SocketOptions.SO_BROADCAST:
                     return new Boolean(CIL.unbox_int(socket.GetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.Broadcast))) != 0);
                 case SocketOptions.IP_MULTICAST_IF:
@@ -472,16 +340,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
         }
     }
 
-    /*************************************************************************/
-
-    /**
-     * Sets the value of an option on the socket
-     *
-     * @param option_id The identifier of the option to set
-     * @param val The value of the option to set
-     *
-     * @exception SocketException If an error occurs
-     */
     public void setOption(int option_id, Object val) throws SocketException
     {
         try
@@ -490,14 +348,11 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
             if(false) throw new cli.System.ObjectDisposedException("");
             switch(option_id)
             {
-                case SocketOptions.SO_REUSEADDR:
-                    socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.ReuseAddress), ((Boolean)val).booleanValue() ? 1 : 0);
-                    break;
                 case SocketOptions.SO_BROADCAST:
                     socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.Broadcast), ((Boolean)val).booleanValue() ? 1 : 0);
                     break;
                 case SocketOptions.IP_MULTICAST_IF:
-                    socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.MulticastInterface), (int)PlainSocketImpl.getAddressFromInetAddress((InetAddress)val));
+                    socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.MulticastInterface), (int)PlainSocketImpl.getAddressFromInetAddress((InetAddress)val).get_Address());
                     break;
                 case SocketOptions.IP_MULTICAST_IF2:
                     throw new SocketException("SocketOptions.IP_MULTICAST_IF2 not implemented");
@@ -506,7 +361,6 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
                     break;
                 case SocketOptions.SO_TIMEOUT:
                     socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.ReceiveTimeout), ((Integer)val).intValue());
-                    socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.Socket), SocketOptionName.wrap(SocketOptionName.SendTimeout), ((Integer)val).intValue());
                     break;
                 default:
                     PlainSocketImpl.setCommonSocketOption(socket, option_id, val);
@@ -526,23 +380,87 @@ public class PlainDatagramSocketImpl extends DatagramSocketImpl
     public int peekData(DatagramPacket packet)
     {
         // TODO
-        throw new InternalError ("PlainDatagramSocketImpl::peekData is not implemented");
+        throw new InternalError("PlainDatagramSocketImpl::peekData is not implemented");
     }
 
-    public void joinGroup(SocketAddress address, NetworkInterface netIf)
+    public void joinGroup(SocketAddress address, NetworkInterface netIf) throws IOException
     {
-        // TODO
-        throw new InternalError ("PlainDatagramSocketImpl::joinGroup is not implemented");
+        try
+        {
+            if(!(address instanceof InetSocketAddress)) throw new cli.System.ArgumentException();
+            if(false) throw new cli.System.Net.Sockets.SocketException();
+            if(false) throw new cli.System.ObjectDisposedException("");
+
+            if(((InetSocketAddress)address).isUnresolved())
+            {
+                throw new UnknownHostException(((InetSocketAddress)address).getHostName());
+            }
+
+            InetAddress inetAddr = ((InetSocketAddress)address).getAddress();
+            IPAddress mcastAddr = PlainSocketImpl.getAddressFromInetAddress(inetAddr);
+
+            Enumeration e = netIf.getInetAddresses();
+            if(e.hasMoreElements())
+            {
+                IPAddress bindAddr  = PlainSocketImpl.getAddressFromInetAddress((InetAddress)e.nextElement());
+                MulticastOption mcastOption = new MulticastOption(mcastAddr, bindAddr);
+                socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.AddMembership), mcastOption);
+            }
+        }
+        catch(cli.System.Net.Sockets.SocketException x)
+        {
+            throw PlainSocketImpl.convertSocketExceptionToIOException(x);
+        }
+        catch(cli.System.ArgumentException x1)
+        {
+            throw new IOException(x1.getMessage());
+        }
+        catch(cli.System.ObjectDisposedException x2)
+        {
+            throw new SocketException("Socket is closed");
+        }
     }
 
-    public void leaveGroup(SocketAddress address, NetworkInterface netIf)
+    public void leaveGroup(SocketAddress address, NetworkInterface netIf) throws IOException
     {
-        // TODO
-        throw new InternalError ("PlainDatagramSocketImpl::leaveGroup is not implemented");
+        try
+        {
+            if(!(address instanceof InetSocketAddress)) throw new cli.System.ArgumentException();
+            if(false) throw new cli.System.Net.Sockets.SocketException();
+            if(false) throw new cli.System.ObjectDisposedException("");
+
+            if(((InetSocketAddress)address).isUnresolved())
+            {
+                throw new UnknownHostException(((InetSocketAddress)address).getHostName());
+            }
+
+            InetAddress inetAddr = ((InetSocketAddress)address).getAddress();
+            IPAddress mcastAddr = PlainSocketImpl.getAddressFromInetAddress(inetAddr);
+
+            Enumeration e = netIf.getInetAddresses();
+            if(e.hasMoreElements())
+            {
+                IPAddress bindAddr  = PlainSocketImpl.getAddressFromInetAddress((InetAddress)e.nextElement());
+                MulticastOption mcastOption = new MulticastOption(mcastAddr, bindAddr);
+                socket.SetSocketOption(SocketOptionLevel.wrap(SocketOptionLevel.IP), SocketOptionName.wrap(SocketOptionName.DropMembership), mcastOption);
+            }
+        }
+        catch(cli.System.Net.Sockets.SocketException x)
+        {
+            throw PlainSocketImpl.convertSocketExceptionToIOException(x);
+        }
+        catch(cli.System.ArgumentException x1)
+        {
+            throw new IOException(x1.getMessage());
+        }
+        catch(cli.System.ObjectDisposedException x2)
+        {
+            throw new SocketException("Socket is closed");
+        }
     }
 
     public cli.System.Net.Sockets.Socket getSocket()
     {
         return socket;
     }
-} // class PlainDatagramSocketImpl
+}
