@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2004 Jeroen Frijters
+  Copyright (C) 2004, 2005, 2006 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -27,7 +27,7 @@ package ikvm.internal;
 public interface LibraryVMInterface
 {
     Object loadClass(Object classLoader, String name) throws ClassNotFoundException;
-    Object newClass(Object wrapper, Object protectionDomain);
+    Object newClass(Object wrapper, Object protectionDomain, Object classLoader);
     Object getWrapperFromClass(Object clazz);
     Object getWrapperFromField(Object field);
     Object getWrapperFromMethodOrConstructor(Object methodOrConstructor);
@@ -35,8 +35,6 @@ public interface LibraryVMInterface
     Object getWrapperFromClassLoader(Object classLoader);
     void setWrapperForClassLoader(Object classLoader, Object wrapper);
 
-    Object getSystemClassLoader();
-
     Object box(Object val);
     Object unbox(Object val);
 
@@ -90,4 +88,6 @@ public interface LibraryVMInterface
     Throwable newInstantiationException(String msg);
     Throwable newInterruptedException();
     Throwable newIllegalMonitorStateException();
+
+    Object newAssemblyClassLoader(cli.System.Reflection.Assembly assembly);
 }
