@@ -1,13 +1,41 @@
+/*
+  Copyright (C) 2003, 2004, 2005, 2006, 2007 Jeroen Frijters
+
+  This software is provided 'as-is', without any express or implied
+  warranty.  In no event will the authors be held liable for any damages
+  arising from the use of this software.
+
+  Permission is granted to anyone to use this software for any purpose,
+  including commercial applications, and to alter it and redistribute it
+  freely, subject to the following restrictions:
+
+  1. The origin of this software must not be misrepresented; you must not
+     claim that you wrote the original software. If you use this software
+     in a product, an acknowledgment in the product documentation would be
+     appreciated but is not required.
+  2. Altered source versions must be plainly marked as such, and must not be
+     misrepresented as being the original software.
+  3. This notice may not be removed or altered from any source distribution.
+
+  Jeroen Frijters
+  jeroen@frijters.net
+  
+*/
 package java.lang;
 
+import ikvm.lang.Internal;
+import java.util.concurrent.atomic.AtomicInteger;
+
 // Note: stop() should take care not to stop a thread if it is
 // executing code in this class.
-final class VMThread
+@Internal
+public final class VMThread
 {
-    private static final Object countLock = new Object();
-    private static int nonDaemonCount;
+    private static final AtomicInteger nonDaemonCount = new AtomicInteger();
     // used by inner class
+    @cli.System.ThreadStaticAttribute.Annotation
     /*private*/ static Thread __tls_javaThread;
+    @cli.System.ThreadStaticAttribute.Annotation
     private static Object __tls_cleanup;
     private cli.System.WeakReference nativeThreadReference;
 
@@ -15,9 +43,13 @@ final class VMThread
     volatile Thread thread;
     private volatile boolean running;
     private volatile boolean interruptableWait;
+    private InterruptProc interruptProc;
+    private boolean timedWait;
     private volatile boolean interruptPending;
     private VMThread firstJoinWaiter;
     private VMThread nextJoinWaiter;
+    private volatile Object blocker;
+    private boolean pendingUnpark;
 
     private VMThread(Thread thread)
     {
@@ -67,20 +99,25 @@ final class VMThread
 
     // notify Thread that it is dead, this method can safely be called multiple times
     // used by inner class
-    /*private*/ synchronized void cleanup()
+    /*private*/ void cleanup()
     {
-	if(thread.vmThread != null)
-	{
-	    thread.die();
-	    nativeThreadReference.set_Target(null);
-	    if(!thread.daemon)
-	    {
-		synchronized(countLock)
-		{
-		    nonDaemonCount--;
-		}
-	    }
-	}
+        // since we're going to call back to Thread.die() (which is synchronized)
+        // we must first acquire the Thread lock and then our own lock, to prevent deadlock.
+        synchronized(thread)
+        {
+            synchronized(this)
+            {
+	        if(thread.vmThread != null)
+	        {
+	            thread.die();
+	            nativeThreadReference.set_Target(null);
+	            if(!thread.daemon)
+	            {
+                        nonDaemonCount.decrementAndGet();
+	            }
+	        }
+            }
+        }
     }
 
     private synchronized void addJoinWaiter(VMThread waiter)
@@ -121,7 +158,7 @@ final class VMThread
         }
     }
 
-    static void jniDetach()
+    public static void jniDetach()
     {
         // If this thread never called Thread.currentThread(), we don't need
         // to clean up, because the Java Thread object doesn't exist and nobody
@@ -150,19 +187,13 @@ final class VMThread
     {
 	if(!Thread.currentThread().isDaemon())
 	{
-	    synchronized(countLock)
-	    {
-		nonDaemonCount--;
-	    }
+            nonDaemonCount.decrementAndGet();
 	}
 	for(;;)
 	{
-	    synchronized(countLock)
+            if(nonDaemonCount.get() == 0)
 	    {
-		if(nonDaemonCount == 0)
-		{
-		    return;
-		}
+                return;
 	    }
 	    try
 	    {
@@ -222,7 +253,7 @@ final class VMThread
 
     void join(long ms, int ns) throws InterruptedException
     {
-	cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+	cli.System.Threading.Thread nativeThread = getNativeThread();
 	if(nativeThread == null)
 	{
 	    return;
@@ -230,7 +261,7 @@ final class VMThread
 	try
 	{
             VMThread current = currentThread().vmThread;
-	    enterInterruptableWait();
+	    enterInterruptableWait(ms != 0 || ns != 0, null);
 	    try
 	    {
                 addJoinWaiter(current);
@@ -291,7 +322,7 @@ final class VMThread
 	    thread.stillborn = t;
     }
 
-    void start(long stacksize)
+    private void start(long stacksize)
     {
 	cli.System.Threading.ThreadStart starter = new cli.System.Threading.ThreadStart(
 	    new cli.System.Threading.ThreadStart.Method()
@@ -310,14 +341,11 @@ final class VMThread
 	nativeThread.Start();
 	if(!thread.daemon)
 	{
-	    synchronized(countLock)
-	    {
-		nonDaemonCount++;
-	    }
+            nonDaemonCount.incrementAndGet();
 	}
     }
 
-    private static void enterInterruptableWait() throws InterruptedException
+    private static void enterInterruptableWait(boolean timedWait, InterruptProc proc) throws InterruptedException
     {
         VMThread vmthread = currentThread().vmThread;
         synchronized(vmthread)
@@ -327,11 +355,24 @@ final class VMThread
                 vmthread.interruptPending = false;
                 throw new InterruptedException();
             }
+            vmthread.timedWait = timedWait;
             vmthread.interruptableWait = true;
+            vmthread.interruptProc = proc;
         }
     }
 
-    private static void leaveInterruptableWait() throws InterruptedException
+    @Internal
+    public static interface InterruptProc
+    {
+        void interrupt();
+    }
+
+    public static void enterInterruptableWait(InterruptProc proc) throws InterruptedException
+    {
+        enterInterruptableWait(false, proc);
+    }
+
+    public static void leaveInterruptableWait() throws InterruptedException
     {
         cli.System.Threading.ThreadInterruptedException dotnetInterrupt = null;
         for(;;)
@@ -342,6 +383,7 @@ final class VMThread
                 VMThread vmthread = currentThread().vmThread;
                 synchronized(vmthread)
                 {
+                    vmthread.interruptProc = null;
                     vmthread.interruptableWait = false;
                     if(vmthread.interruptPending)
                     {
@@ -367,11 +409,18 @@ final class VMThread
         interruptPending = true;
         if(interruptableWait)
         {
-            cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
-	    if(nativeThread != null)
-	    {
-	        nativeThread.Interrupt();
-	    }
+            if(interruptProc != null)
+            {
+                interruptProc.interrupt();
+            }
+            else
+            {
+                cli.System.Threading.Thread nativeThread = getNativeThread();
+	        if(nativeThread != null)
+	        {
+	            nativeThread.Interrupt();
+	        }
+            }
         }
     }
 
@@ -393,7 +442,7 @@ final class VMThread
 
     void suspend()
     {
-	cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+	cli.System.Threading.Thread nativeThread = getNativeThread();
 	if(nativeThread != null)
 	{
             try
@@ -409,7 +458,7 @@ final class VMThread
 
     void resume()
     {
-	cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+	cli.System.Threading.Thread nativeThread = getNativeThread();
 	if(nativeThread != null)
 	{
             try
@@ -425,7 +474,7 @@ final class VMThread
 
     void nativeSetPriority(int priority)
     {
-	cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+	cli.System.Threading.Thread nativeThread = getNativeThread();
 	if(nativeThread != null)
 	{
 	    if(priority == Thread.MIN_PRIORITY)
@@ -469,7 +518,7 @@ final class VMThread
 	}
 	else if(t instanceof ThreadDeath)
 	{
-	    cli.System.Threading.Thread nativeThread = (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+	    cli.System.Threading.Thread nativeThread = getNativeThread();
 	    if(nativeThread != null)
 	    {
                 try
@@ -520,6 +569,14 @@ final class VMThread
     // of Java (either in .NET or in native code)
     private static Thread newThread(ThreadGroup group)
     {
+        // protect against calls to Thread.currentThread() while we're creating the Thread object
+        if(__tls_cleanup != null)
+        {
+            throw new InternalError("Recursive Thread.currentThread() initialization");
+        }
+        // set a temporary value, so that we know we're initializing this thread
+        __tls_cleanup = "";
+
         cli.System.Threading.Thread nativeThread = cli.System.Threading.Thread.get_CurrentThread();
         VMThread vmThread = new VMThread(null);
         vmThread.nativeThreadReference = new cli.System.WeakReference(nativeThread);
@@ -546,10 +603,7 @@ final class VMThread
         Thread javaThread = new Thread(vmThread, nativeThread.get_Name(), priority, nativeThread.get_IsBackground());
         if(!javaThread.daemon)
         {
-            synchronized(countLock)
-            {
-                nonDaemonCount++;
-            }
+            nonDaemonCount.incrementAndGet();
         }
         vmThread.thread = javaThread;
         __tls_javaThread = javaThread;
@@ -565,7 +619,7 @@ final class VMThread
         Thread javaThread = __tls_javaThread;
 	if(javaThread == null)
 	{
-            __tls_javaThread = javaThread = newThread(ThreadGroup.root);
+            javaThread = newThread(ThreadGroup.root);
 	}
 	return javaThread;
     }
@@ -587,7 +641,7 @@ final class VMThread
 	}
 	else
 	{
-            enterInterruptableWait();
+            enterInterruptableWait(true, null);
             try
             {
 	        // if nanoseconds are specified, round up to one millisecond
@@ -639,7 +693,7 @@ final class VMThread
         {
             throw new IllegalArgumentException("argument out of range");
         }
-        enterInterruptableWait();
+        enterInterruptableWait(timeout != 0 || nanos != 0, null);
         try
         {
             if((timeout == 0 && nanos == 0) || timeout > 922337203685476L)
@@ -656,4 +710,113 @@ final class VMThread
             leaveInterruptableWait();
         }
     }
+
+    public static void park(Object blocker, long nanos)
+    {
+        VMThread vmthread = currentThread().vmThread;
+        vmthread.blocker = blocker;
+        try
+        {
+            vmthread.parkImpl(nanos);
+        }
+        finally
+        {
+            vmthread.blocker = null;
+        }
+    }
+
+    public static void park(long nanos)
+    {
+        currentThread().vmThread.parkImpl(nanos);
+    }
+
+    private synchronized void parkImpl(long nanos)
+    {
+        if(nanos > 0 && !interruptPending)
+        {
+            long millis = nanos / 1000000;
+            if(millis > Integer.MAX_VALUE)
+            {
+                millis = Integer.MAX_VALUE;
+            }
+            if(!pendingUnpark)
+            {
+                timedWait = millis != Integer.MAX_VALUE;
+                interruptableWait = true;
+                try
+                {
+                    if(false) throw new cli.System.Threading.ThreadInterruptedException();
+                    cli.System.Threading.Monitor.Wait(this, (int)millis);
+                }
+                catch(cli.System.Threading.ThreadInterruptedException x)
+                {
+                }
+                finally
+                {
+                    interruptableWait = false;
+                }
+            }
+            pendingUnpark = false;
+        }
+    }
+
+    public static void unpark(Thread t)
+    {
+        if(t != null)
+        {
+            // NOTE we don't queue an unpark if the thread is not yet started
+            // (Sun doesn't either)
+            VMThread vmthread = t.vmThread;
+            if(vmthread != null)
+            {
+                synchronized(vmthread)
+                {
+                    vmthread.pendingUnpark = true;
+                    cli.System.Threading.Monitor.Pulse(vmthread);
+                }
+            }
+        }
+    }
+
+    public static Object getBlocker(Thread t)
+    {
+        VMThread vmthread = t.vmThread;
+        return vmthread != null ? vmthread.blocker : null;
+    }
+
+    synchronized String getState()
+    {
+        cli.System.Threading.Thread nativeThread = getNativeThread();
+        if(nativeThread != null && nativeThread.get_IsAlive())
+        {
+            if(interruptableWait)
+            {
+                // NOTE if objectWait has satisfied the wait condition (or has been interrupted or has timed-out),
+                // it can be blocking on the re-acquire of the monitor, but we have no way of detecting that.
+                return timedWait ? "TIMED_WAITING" : "WAITING";
+            }
+            if((nativeThread.get_ThreadState().Value & cli.System.Threading.ThreadState.WaitSleepJoin) != 0)
+            {
+                return "BLOCKED";
+            }
+            return "RUNNABLE";
+        }
+        return "TERMINATED";
+    }
+
+    private cli.System.Threading.Thread getNativeThread()
+    {
+        return (cli.System.Threading.Thread)nativeThreadReference.get_Target();
+    }
+
+    public static cli.System.Threading.Thread getNativeThread(Thread t)
+    {
+        VMThread vmthread = t.vmThread;
+        return vmthread != null ? vmthread.getNativeThread() : null;
+    }
+
+    public static Thread getThreadFromId(long id)
+    {
+        return ThreadGroup.getThreadFromId(id);
+    }
 }
