@@ -29,8 +29,8 @@ using System.Collections;
 using System.Collections.Specialized;
 using System.Text;
 using IKVM.Internal;
-using IKVM.Runtime;
 
+using ikvm.runtime;
 using java.lang.reflect;
 using java.net;
 using java.util.jar;
@@ -54,36 +54,6 @@ public class Starter
 		}
 	}
 
-	public class PathClassLoader : URLClassLoader
-	{
-		private static URL[] GetClassPath(string classpath)
-		{
-			string[] s = classpath.Split(Path.PathSeparator);
-			URL[] urls = new URL[s.Length];
-			for(int i = 0; i < urls.Length; i++)
-			{
-				// TODO non-existing file/dir is treated as current directory, this obviously isn't correct
-				urls[i] = new java.io.File(s[i]).toURL();
-			}
-			return urls;
-		}
-
-		public PathClassLoader(string classpath, java.lang.ClassLoader parent)
-			: base(GetClassPath(classpath), parent)
-		{
-		}
-
-		protected override java.lang.Class loadClass(string name, bool resolve)
-		{
-			java.lang.Class c = findClass(name);
-			if(resolve)
-			{
-				resolveClass(c);
-			}
-			return c;
-		}
-	}
-
 	private class SaveAssemblyShutdownHook : java.lang.Thread
 	{
 		private java.lang.Class clazz;
@@ -100,7 +70,7 @@ public class Starter
 			Console.Error.WriteLine("Saving dynamic assembly...");
 			try
 			{
-				IKVM.Internal.Starter.SaveDebugImage(clazz);
+				IKVM.Internal.Starter.SaveDebugImage();
 				Console.Error.WriteLine("Saving done.");
 			}
 			catch(Exception x)
@@ -135,6 +105,12 @@ public class Starter
 	{
 		Tracer.EnableTraceForDebug();
 		Hashtable props = new Hashtable();
+		string classpath = Environment.GetEnvironmentVariable("CLASSPATH");
+		if(classpath == null || classpath == "")
+		{
+			classpath = ".";
+		}
+		props["java.class.path"] = classpath;
 		bool jar = false;
 		bool saveAssembly = false;
 		bool saveAssemblyX = false;
@@ -142,7 +118,6 @@ public class Starter
 		bool showVersion = false;
 		string mainClass = null;
 		string[] vmargs = null;
-		string bootclasspath = null;
 		for(int i = 0; i < args.Length; i++)
 		{
 			if(args[i][0] == '-')
@@ -179,9 +154,14 @@ public class Starter
 				}
 				else if(args[i] == "-version")
 				{
-					Console.WriteLine(IKVM.Runtime.Startup.GetVersionAndCopyrightInfo());
+					Console.WriteLine(Startup.getVersionAndCopyrightInfo());
 					Console.WriteLine();
 					Console.WriteLine("CLR version: {0} ({1} bit)", Environment.Version, IntPtr.Size * 8);
+					System.Type type = System.Type.GetType("Mono.Runtime");
+					if(type != null)
+					{
+						Console.WriteLine("Mono version: {0}", type.InvokeMember("GetDisplayName", BindingFlags.InvokeMethod | BindingFlags.Static | BindingFlags.NonPublic, null, null, new object[0]));
+					}
 					foreach(Assembly asm in AppDomain.CurrentDomain.GetAssemblies())
 					{
 						Console.WriteLine("{0}: {1}", asm.GetName().Name, asm.GetName().Version);
@@ -222,10 +202,6 @@ public class Starter
 				{
 					props["java.class.path"] = args[++i];
 				}
-				else if(args[i].StartsWith("-Xbootclasspath:"))
-				{
-					bootclasspath = args[i].Substring(16);
-				}
 				else if(args[i].StartsWith("-Xtrace:"))
 				{
 					Tracer.SetTraceLevel(args[i].Substring(8));
@@ -261,13 +237,13 @@ public class Starter
 			else
 			{
 				mainClass = args[i];
-				vmargs = Startup.Glob(i + 2);
+				vmargs = Startup.glob(i + 2);
 				break;
 			}
 		}
 		if(mainClass == null || showVersion)
 		{
-			Console.Error.WriteLine(IKVM.Runtime.Startup.GetVersionAndCopyrightInfo());
+			Console.Error.WriteLine(Startup.getVersionAndCopyrightInfo());
 			Console.Error.WriteLine();
 		}
 		if(mainClass == null)
@@ -292,8 +268,6 @@ public class Starter
 			Console.Error.WriteLine("                      Disable assertions");
 			Console.Error.WriteLine("    -Xsave            Save the generated assembly (for debugging)");
 			Console.Error.WriteLine("    -Xtime            Time the execution");
-			Console.Error.WriteLine("    -Xbootclasspath:<directories and zip/jar files separated by {0}>", Path.PathSeparator);
-			Console.Error.WriteLine("                      Set search path for bootstrap classes and resources");
 			Console.Error.WriteLine("    -Xtrace:<string>  Displays all tracepoints with the given name");
 			Console.Error.WriteLine("    -Xmethodtrace:<string>");
 			Console.Error.WriteLine("                      Builds method trace into the specified output methods");
@@ -307,8 +281,8 @@ public class Starter
 			{
 				props["java.class.path"] = mainClass;
 			}
-			Startup.SetProperties(props);
-			Startup.EnterMainThread();
+			Startup.setProperties(props);
+			Startup.enterMainThread();
 			if(jar)
 			{
 				JarFile jf = new JarFile(mainClass);
@@ -333,58 +307,57 @@ public class Starter
 				}
 				mainClass = mainClass.Replace('/', '.');
 			}
-			if(bootclasspath != null)
-			{
-				IKVM.Internal.Starter.SetBootstrapClassLoader(new PathClassLoader(bootclasspath, null));
-			}
 			java.lang.Class clazz = java.lang.Class.forName(mainClass, true, java.lang.ClassLoader.getSystemClassLoader());
-			Method method = FindMainMethod(clazz);
-			if(method == null)
+			try
 			{
-				throw new java.lang.NoSuchMethodError("main");
-			}
-			else if(!Modifier.isPublic(method.getModifiers()))
-			{
-				Console.Error.WriteLine("Main method not public.");
-			}
-			else
-			{
-				// if clazz isn't public, we can still call main
-				method.setAccessible(true);
-				if(saveAssembly)
+				Method method = FindMainMethod(clazz);
+				if(method == null)
 				{
-					java.lang.Runtime.getRuntime().addShutdownHook(new SaveAssemblyShutdownHook(clazz));
+					throw new java.lang.NoSuchMethodError("main");
 				}
-				if(waitOnExit)
+				else if(!Modifier.isPublic(method.getModifiers()))
 				{
-					java.lang.Runtime.getRuntime().addShutdownHook(new WaitShutdownHook());
+					Console.Error.WriteLine("Main method not public.");
 				}
-				try
+				else
 				{
-					method.invoke(null, new object[] { vmargs });
-					return 0;
-				}
-				catch(InvocationTargetException x)
-				{
-					throw x.getCause();
-				}
-				finally
-				{
-					if(saveAssemblyX)
+					// if clazz isn't public, we can still call main
+					method.setAccessible(true);
+					if(saveAssembly)
 					{
-						IKVM.Internal.Starter.SaveDebugImage(clazz);
+						java.lang.Runtime.getRuntime().addShutdownHook(new SaveAssemblyShutdownHook(clazz));
 					}
+					if(waitOnExit)
+					{
+						java.lang.Runtime.getRuntime().addShutdownHook(new WaitShutdownHook());
+					}
+					try
+					{
+						method.invoke(null, new object[] { vmargs });
+						return 0;
+					}
+					catch(InvocationTargetException x)
+					{
+						throw x.getCause();
+					}
+				}
+			}
+			finally
+			{
+				if(saveAssemblyX)
+				{
+					IKVM.Internal.Starter.SaveDebugImage();
 				}
 			}
 		}
 		catch(System.Exception x)
 		{
 			java.lang.Thread thread = java.lang.Thread.currentThread();
-			thread.getThreadGroup().uncaughtException(thread, IKVM.Runtime.Util.MapException(x));
+			thread.getThreadGroup().uncaughtException(thread, ikvm.runtime.Util.mapException(x));
 		}
 		finally
 		{
-			Startup.ExitMainThread();
+			Startup.exitMainThread();
 		}
 		return 1;
 	}
