@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002, 2003, 2004, 2005, 2006 Jeroen Frijters
+  Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -33,9 +33,9 @@ using IKVM.Internal;
 class IkvmcCompiler
 {
 	private static string manifestMainClass;
-	private static ArrayList classes = new ArrayList();
+	private static Hashtable classes = new Hashtable();
 	private static Hashtable resources = new Hashtable();
-	private static bool compressResources;
+	private static bool time;
 
 	private static ArrayList GetArgs(string[] args)
 	{
@@ -63,9 +63,35 @@ class IkvmcCompiler
 
 	static void Main(string[] args)
 	{
+		DateTime start = DateTime.Now;
+		int rc = RealMain(args);
+		if (time)
+		{
+			Console.WriteLine("Total cpu time: {0}", System.Diagnostics.Process.GetCurrentProcess().TotalProcessorTime);
+			Console.WriteLine("Total wall clock time: {0}", DateTime.Now - start);
+		}
 		// FXBUG if we run a static initializer that starts a thread, we would never end,
 		// so we force an exit here
-		Environment.Exit(RealMain(args));
+		Environment.Exit(rc);
+	}
+
+	static string GetVersionAndCopyrightInfo()
+	{
+		Assembly asm = Assembly.GetEntryAssembly();
+		object[] desc = asm.GetCustomAttributes(typeof(AssemblyTitleAttribute), false);
+		if (desc.Length == 1)
+		{
+			object[] copyright = asm.GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false);
+			if (copyright.Length == 1)
+			{
+				return string.Format("{0} version {1}{2}{3}{2}http://www.ikvm.net/",
+					((AssemblyTitleAttribute)desc[0]).Title,
+					asm.GetName().Version,
+					Environment.NewLine,
+					((AssemblyCopyrightAttribute)copyright[0]).Copyright);
+			}
+		}
+		return "";
 	}
 
 	static int RealMain(string[] args)
@@ -89,11 +115,12 @@ class IkvmcCompiler
 		options.props = new Hashtable();
 		if(arglist.Count == 0)
 		{
-			Console.Error.WriteLine(IKVM.Runtime.Startup.GetVersionAndCopyrightInfo());
+			Console.Error.WriteLine(GetVersionAndCopyrightInfo());
 			Console.Error.WriteLine();
 			Console.Error.WriteLine("usage: ikvmc [-options] <classOrJar1> ... <classOrJarN>");
 			Console.Error.WriteLine();
 			Console.Error.WriteLine("options:");
+			Console.Error.WriteLine("    @<filename>                Read more options from file");
 			Console.Error.WriteLine("    -out:<outputfile>          Specify the output filename");
 			Console.Error.WriteLine("    -assembly:<name>           Specify assembly name");
 			Console.Error.WriteLine("    -target:exe                Build a console executable");
@@ -109,6 +136,8 @@ class IkvmcCompiler
 			Console.Error.WriteLine("    -recurse:<filespec>        Recurse directory and include matching files");
 			Console.Error.WriteLine("    -nojni                     Do not generate JNI stub for native methods");
 			Console.Error.WriteLine("    -resource:<name>=<path>    Include file as Java resource");
+			Console.Error.WriteLine("    -externalresource:<name>=<path>");
+			Console.Error.WriteLine("                               Reference file as Java resource");
 			Console.Error.WriteLine("    -exclude:<filename>        A file containing a list of classes to exclude");
 			Console.Error.WriteLine("    -debug                     Generate debug info for the output file");
 			Console.Error.WriteLine("                               (Note that this also causes the compiler to");
@@ -134,6 +163,9 @@ class IkvmcCompiler
 			Console.Error.WriteLine("                               of initializer methods");
 			Console.Error.WriteLine("    -privatepackage:<prefix>   Mark all classes with a package name starting");
 			Console.Error.WriteLine("                               with <prefix> as internal to the assembly");
+			Console.Error.WriteLine("    -nowarn:<warning[:key]>    Suppress specified warnings");
+			Console.Error.WriteLine("    -warnaserror:<warning[:key]>  Treat specified warnings as errors");
+			Console.Error.WriteLine("    -time                      Display timing statistics");
 			return 1;
 		}
 		foreach(string s in arglist)
@@ -240,8 +272,13 @@ class IkvmcCompiler
 					string[] files = Directory.GetFiles(path == "" ? "." : path, Path.GetFileName(r));
 					if(files.Length == 0)
 					{
-						Console.Error.WriteLine("Error: reference not found: {0}", r);
-						return 1;
+						Assembly asm = Assembly.LoadWithPartialName(r);
+						if(asm == null)
+						{
+							Console.Error.WriteLine("Error: reference not found: {0}", r);
+							return 1;
+						}
+						files = new string[] { asm.Location };
 					}
 					foreach(string f in files)
 					{
@@ -277,6 +314,11 @@ class IkvmcCompiler
 							Console.Error.WriteLine("Error: path too long: {0}", spec);
 							return 1;
 						}
+						catch(DirectoryNotFoundException)
+						{
+							Console.Error.WriteLine("Error: path not found: {0}", spec);
+							return 1;
+						}
 						catch(ArgumentException)
 						{
 							Console.Error.WriteLine("Error: invalid path: {0}", spec);
@@ -287,31 +329,50 @@ class IkvmcCompiler
 				else if(s.StartsWith("-resource:"))
 				{
 					string[] spec = s.Substring(10).Split('=');
-					if(resources.ContainsKey(spec[0]))
+					try
 					{
-						Console.Error.WriteLine("Warning: skipping resource (name clash): " + spec[0]);
-					}
-					else
-					{
-						try
+						using(FileStream fs = new FileStream(spec[1], FileMode.Open, FileAccess.Read))
 						{
-							using(FileStream fs = new FileStream(spec[1], FileMode.Open, FileAccess.Read))
+							byte[] b = new byte[fs.Length];
+							fs.Read(b, 0, b.Length);
+							string name = spec[0];
+							if(name.StartsWith("/"))
 							{
-								byte[] b = new byte[fs.Length];
-								fs.Read(b, 0, b.Length);
-								AddResource(spec[0], b);
+								// a leading slash is not required, so strip it
+								name = name.Substring(1);
 							}
-						}
-						catch(Exception x)
-						{
-							Console.Error.WriteLine("Error: {0}: {1}", x.Message, spec[1]);
-							return 1;
+							AddResource(name, b);
 						}
 					}
+					catch(Exception x)
+					{
+						Console.Error.WriteLine("Error: {0}: {1}", x.Message, spec[1]);
+						return 1;
+					}
+				}
+				else if(s.StartsWith("-externalresource:"))
+				{
+					string[] spec = s.Substring(18).Split('=');
+					if(!File.Exists(spec[1]))
+					{
+						Console.Error.WriteLine("Error: external resource file does not exist: {0}", spec[1]);
+						return 1;
+					}
+					if(Path.GetFileName(spec[1]) != spec[1])
+					{
+						Console.Error.WriteLine("Error: external resource file may not include path specification: {0}", spec[1]);
+						return 1;
+					}
+					if(options.externalResources == null)
+					{
+						options.externalResources = new Hashtable();
+					}
+					// TODO resource name clashes should be tested
+					options.externalResources.Add(spec[0], spec[1]);
 				}
 				else if(s == "-nojni")
 				{
-					options.nojni = true;
+					options.codegenoptions |= CodeGenOptions.NoJNI;
 				}
 				else if(s.StartsWith("-exclude:"))
 				{
@@ -359,11 +420,11 @@ class IkvmcCompiler
 				}
 				else if(s == "-debug")
 				{
-					JVM.Debug = true;
+					options.codegenoptions |= CodeGenOptions.Debug;
 				}
 				else if(s.StartsWith("-srcpath:"))
 				{
-					JVM.SourcePath = s.Substring(9);
+					options.sourcepath = s.Substring(9);
 				}
 				else if(s.StartsWith("-remap:"))
 				{
@@ -371,7 +432,7 @@ class IkvmcCompiler
 				}
 				else if(s == "-nostacktraceinfo")
 				{
-					options.nostacktraceinfo = true;
+					options.codegenoptions |= CodeGenOptions.NoStackTraceInfo;
 				}
 				else if(s == "-opt:fields")
 				{
@@ -379,11 +440,11 @@ class IkvmcCompiler
 				}
 				else if(s == "-compressresources")
 				{
-					compressResources = true;
+					options.compressedResources = true;
 				}
 				else if(s == "-strictfinalfieldsemantics")
 				{
-					options.strictFinalFieldSemantics = true;
+					options.codegenoptions |= CodeGenOptions.StrictFinalFieldSemantics;
 				}
 				else if(s.StartsWith("-privatepackage:"))
 				{
@@ -400,11 +461,41 @@ class IkvmcCompiler
 						options.privatePackages = temp;
 					}
 				}
+				else if(s.StartsWith("-nowarn:"))
+				{
+					foreach(string w in s.Substring(8).Split(','))
+					{
+						string ws = w;
+						// lame way to chop off the leading zeroes
+						while(ws.StartsWith("0"))
+						{
+							ws = ws.Substring(1);
+						}
+						StaticCompiler.SuppressWarning(ws);
+					}
+				}
+				else if(s.StartsWith("-warnaserror:"))
+				{
+					foreach(string w in s.Substring(13).Split(','))
+					{
+						string ws = w;
+						// lame way to chop off the leading zeroes
+						while(ws.StartsWith("0"))
+						{
+							ws = ws.Substring(1);
+						}
+						StaticCompiler.WarnAsError(ws);
+					}
+				}
 				else if(s.StartsWith("-runtime:"))
 				{
 					// NOTE this is an undocumented option
 					options.runtimeAssembly = s.Substring(9);
 				}
+				else if(s == "-time")
+				{
+					time = true;
+				}
 				else
 				{
 					Console.Error.WriteLine("Warning: unrecognized option: {0}", s);
@@ -446,7 +537,6 @@ class IkvmcCompiler
 				}
 			}
 		}
-		options.compressedResources = compressResources;
 		if(options.assembly == null)
 		{
 			string basename = options.path == null ? defaultAssemblyName : new FileInfo(options.path).Name;
@@ -475,16 +565,16 @@ class IkvmcCompiler
 		}
 		if(options.mainClass == null && manifestMainClass != null && (options.guessFileKind || options.target != System.Reflection.Emit.PEFileKinds.Dll))
 		{
-			Console.Error.WriteLine("Note: using main class {0} based on jar manifest", manifestMainClass);
+			StaticCompiler.IssueMessage(Message.MainMethodFromManifest, manifestMainClass);
 			options.mainClass = manifestMainClass;
 		}
 		try
 		{
-			options.classes = (byte[][])classes.ToArray(typeof(byte[]));
+			options.classes = classes;
 			options.references = (string[])references.ToArray(typeof(string));
 			options.resources = resources;
 			options.classesToExclude = (string[])classesToExclude.ToArray(typeof(string));
-			return AotCompiler.Compile(options);
+			return CompilerClassLoader.Compile(options);
 		}
 		catch(Exception x)
 		{
@@ -505,6 +595,36 @@ class IkvmcCompiler
 		return buf;
 	}
 
+	private static void AddClassFile(string filename, byte[] buf, bool addResourceFallback)
+	{
+		try
+		{
+			string name = ClassFile.GetClassName(buf, 0, buf.Length);
+			if(classes.ContainsKey(name))
+			{
+				StaticCompiler.IssueMessage(Message.DuplicateClassName, name);
+			}
+			else
+			{
+				classes.Add(name, buf);
+			}
+		}
+		catch(ClassFormatError x)
+		{
+			if(addResourceFallback)
+			{
+				// not a class file, so we include it as a resource
+				// (IBM's db2os390/sqlj jars apparantly contain such files)
+				StaticCompiler.IssueMessage(Message.NotAClassFile, filename, x.Message);
+				AddResource(filename, buf);
+			}
+			else
+			{
+				StaticCompiler.IssueMessage(Message.ClassFormatError, filename, x.Message);
+			}
+		}
+	}
+
 	private static void ProcessZipFile(string file)
 	{
 		ZipFile zf = new ZipFile(file);
@@ -518,7 +638,7 @@ class IkvmcCompiler
 				}
 				else if(ze.Name.ToLower().EndsWith(".class"))
 				{
-					classes.Add(ReadFromZip(zf, ze));
+					AddClassFile(ze.Name, ReadFromZip(zf, ze), true);
 				}
 				else
 				{
@@ -539,14 +659,7 @@ class IkvmcCompiler
 							}
 						}
 					}
-					if(resources.ContainsKey(ze.Name))
-					{
-						Console.Error.WriteLine("Warning: skipping resource (name clash): " + ze.Name);
-					}
-					else
-					{
-						AddResource(ze.Name, ReadFromZip(zf, ze));
-					}
+					AddResource(ze.Name, ReadFromZip(zf, ze));
 				}
 			}
 		}
@@ -558,17 +671,14 @@ class IkvmcCompiler
 
 	private static void AddResource(string name, byte[] buf)
 	{
-#if !WHIDBEY
-		if(compressResources)
+		if(resources.ContainsKey(name))
 		{
-			MemoryStream mem = new MemoryStream();
-			LZOutputStream lz = new LZOutputStream(mem);
-			lz.Write(buf, 0, buf.Length);
-			lz.Flush();
-			buf = mem.ToArray();
+			StaticCompiler.IssueMessage(Message.DuplicateResourceName, name);
+		}
+		else
+		{
+			resources.Add(name, buf);
 		}
-#endif
-		resources.Add(name, buf);
 	}
 
 	private static void ProcessFile(DirectoryInfo baseDir, string file)
@@ -578,9 +688,9 @@ class IkvmcCompiler
 			case ".class":
 				using(FileStream fs = new FileStream(file, FileMode.Open, FileAccess.Read))
 				{
-					byte[] b = new byte[fs.Length];
-					fs.Read(b, 0, b.Length);
-					classes.Add(b);
+					byte[] buf = new byte[fs.Length];
+					fs.Read(buf, 0, buf.Length);
+					AddClassFile(file, buf, false);
 				}
 				break;
 			case ".jar":
@@ -680,7 +790,11 @@ class IkvmcCompiler
 				return a;
 			}
 		}
+#if WHIDBEY
+		return Assembly.ReflectionOnlyLoad(args.Name);
+#else
 		return null;
+#endif
 	}
 }
 
