@@ -211,6 +211,19 @@ namespace IKVM.Internal
 			return type;
 		}
 
+		internal override Type GetType(string name)
+		{
+			foreach(AssemblyClassLoader loader in referencedAssemblies)
+			{
+				Type type = loader.GetType(name);
+				if(type != null)
+				{
+					return type;
+				}
+			}
+			return null;
+		}
+
 		internal void SetMain(MethodInfo m, PEFileKinds target, Hashtable props, bool noglobbing, Type apartmentAttributeType)
 		{
 			Type[] args = Type.EmptyTypes;
@@ -2183,6 +2196,7 @@ namespace IKVM.Internal
 				Assembly.ReflectionOnlyLoadFrom(typeof(System.ComponentModel.EditorBrowsableAttribute).Assembly.Location);
 			}
 #endif
+			ArrayList assemblyAnnotations = new ArrayList();
 			Hashtable baseClasses = new Hashtable();
 			Hashtable h = new Hashtable();
 			Tracer.Info(Tracer.Compiler, "Parsing class files");
@@ -2197,6 +2211,12 @@ namespace IKVM.Internal
 					{
 						baseClasses[f.SuperClass] = f.SuperClass;
 					}
+					// NOTE the "assembly" type in the unnamed package is a magic type
+					// that acts as the placeholder for assembly attributes
+					if(f.Name == "assembly" && f.Annotations != null)
+					{
+						assemblyAnnotations.AddRange(f.Annotations);
+					}
 				}
 				catch(ClassFormatError)
 				{
@@ -2486,6 +2506,14 @@ namespace IKVM.Internal
 				CustomAttributeBuilder filever = new CustomAttributeBuilder(typeof(AssemblyFileVersionAttribute).GetConstructor(new Type[] { typeof(string) }), new object[] { options.fileversion });
 				loader.assemblyBuilder.SetCustomAttribute(filever);
 			}
+			foreach(object[] def in assemblyAnnotations)
+			{
+				Annotation annotation = Annotation.Load(loader, def);
+				if(annotation != null)
+				{
+					annotation.Apply(loader.assemblyBuilder, def);
+				}
+			}
 			loader.assemblyBuilder.DefineVersionInfoResource();
 			loader.Save();
 			return 0;
