@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002, 2003, 2004, 2005 Jeroen Frijters
+  Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -436,6 +436,24 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("beq")]
+	public sealed class Beq : Branch
+	{
+		public Beq()
+			: base(OpCodes.Beq)
+		{
+		}
+	}
+
+	[XmlType("bne_un")]
+	public sealed class Bne_Un : Branch
+	{
+		public Bne_Un()
+			: base(OpCodes.Bne_Un)
+		{
+		}
+	}
+
 	[XmlType("bge_un")]
 	public sealed class Bge_Un : Branch
 	{
@@ -631,6 +649,14 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("ldind_ref")]
+	public sealed class Ldind_ref : Simple
+	{
+		public Ldind_ref() : base(OpCodes.Ldind_Ref)
+		{
+		}
+	}
+
 	[XmlType("stind_i1")]
 	public sealed class Stind_i1 : Simple
 	{
@@ -639,6 +665,30 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("stind_i2")]
+	public sealed class Stind_i2 : Simple
+	{
+		public Stind_i2() : base(OpCodes.Stind_I2)
+		{
+		}
+	}
+
+	[XmlType("stind_i4")]
+	public sealed class Stind_i4 : Simple
+	{
+		public Stind_i4() : base(OpCodes.Stind_I4)
+		{
+		}
+	}
+
+	[XmlType("stind_ref")]
+	public sealed class Stind_ref : Simple
+	{
+		public Stind_ref() : base(OpCodes.Stind_Ref)
+		{
+		}
+	}
+
 	[XmlType("ret")]
 	public sealed class Ret : Simple
 	{
@@ -696,6 +746,8 @@ namespace IKVM.Internal.MapXml
 	{
 		[XmlAttribute("class")]
 		public string Class;
+		[XmlAttribute("type")]
+		public string Type;
 		[XmlAttribute("name")]
 		public string Name;
 		[XmlAttribute("sig")]
@@ -703,9 +755,16 @@ namespace IKVM.Internal.MapXml
 
 		internal override void Generate(Hashtable context, ILGenerator ilgen)
 		{
-			FieldWrapper fw = ClassLoaderWrapper.LoadClassCritical(Class).GetFieldWrapper(Name, Sig);
-			fw.Link();
-			fw.EmitGet(ilgen);
+			if(Type != null)
+			{
+				ilgen.Emit(OpCodes.Ldsfld, StaticCompiler.GetType(Type).GetField(Name, BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic));
+			}
+			else
+			{
+				FieldWrapper fw = ClassLoaderWrapper.LoadClassCritical(Class).GetFieldWrapper(Name, Sig);
+				fw.Link();
+				fw.EmitGet(ilgen);
+			}
 		}
 	}
 
@@ -871,6 +930,15 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("sub")]
+	public sealed class Sub : Simple
+	{
+		public Sub()
+			: base(OpCodes.Sub)
+		{
+		}
+	}
+
 	[XmlType("mul")]
 	public sealed class Mul : Simple
 	{
@@ -973,6 +1041,30 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("ldelema")]
+	public sealed class Ldelema : Instruction
+	{
+		[XmlAttribute("sig")]
+		public string Sig;
+
+		internal override void Generate(Hashtable context, ILGenerator ilgen)
+		{
+			ilgen.Emit(OpCodes.Ldelema, ClassLoaderWrapper.GetBootstrapClassLoader().FieldTypeWrapperFromSig(Sig).TypeAsArrayType);
+		}
+	}
+
+	[XmlType("ldtoken")]
+	public sealed class Ldtoken : Instruction
+	{
+		[XmlAttribute("type")]
+		public string type;
+
+		internal override void Generate(Hashtable context, ILGenerator ilgen)
+		{
+			ilgen.Emit(OpCodes.Ldtoken, StaticCompiler.GetType(type));
+		}
+	}
+
 	public class InstructionList : CodeEmitter
 	{
 		[XmlElement(typeof(Ldstr))]
@@ -988,6 +1080,8 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(BrFalse))]
 		[XmlElement(typeof(BrTrue))]
 		[XmlElement(typeof(Br))]
+		[XmlElement(typeof(Beq))]
+		[XmlElement(typeof(Bne_Un))]
 		[XmlElement(typeof(Bge_Un))]
 		[XmlElement(typeof(Ble_Un))]
 		[XmlElement(typeof(Blt))]
@@ -1007,7 +1101,11 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(Ldind_i8))]
 		[XmlElement(typeof(Ldind_r4))]
 		[XmlElement(typeof(Ldind_r8))]
+		[XmlElement(typeof(Ldind_ref))]
 		[XmlElement(typeof(Stind_i1))]
+		[XmlElement(typeof(Stind_i2))]
+		[XmlElement(typeof(Stind_i4))]
+		[XmlElement(typeof(Stind_ref))]
 		[XmlElement(typeof(Ret))]
 		[XmlElement(typeof(Throw))]
 		[XmlElement(typeof(Ldnull))]
@@ -1032,12 +1130,15 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(Ldlen))]
 		[XmlElement(typeof(ExceptionBlock))]
 		[XmlElement(typeof(Add))]
+		[XmlElement(typeof(Sub))]
 		[XmlElement(typeof(Mul))]
 		[XmlElement(typeof(Unaligned))]
 		[XmlElement(typeof(Cpblk))]
 		[XmlElement(typeof(Ceq))]
 		[XmlElement(typeof(ConditionalInstruction))]
 		[XmlElement(typeof(Volatile))]
+		[XmlElement(typeof(Ldelema))]
+		[XmlElement(typeof(Ldtoken))]
 		public Instruction[] invoke;
 
 		internal void Generate(Hashtable context, ILGenerator ilgen)
@@ -1104,6 +1205,43 @@ namespace IKVM.Internal.MapXml
 		public string Sig;
 		[XmlAttribute("type")]
 		public string Type;
+
+		internal void Emit(ILGenerator ilgen)
+		{
+			if(Type != "static" || Class == null || Name == null || Sig == null)
+			{
+				throw new NotImplementedException();
+			}
+			Type[] redirParamTypes = ClassLoaderWrapper.GetBootstrapClassLoader().ArgTypeListFromSig(Sig);
+			for(int i = 0; i < redirParamTypes.Length; i++)
+			{
+				ilgen.Emit(OpCodes.Ldarg, (short)i);
+			}
+			// HACK if the class name contains a comma, we assume it is a .NET type
+			if(Class.IndexOf(',') >= 0)
+			{
+				Type type = System.Type.GetType(Class, true);
+				MethodInfo mi = type.GetMethod(Name, redirParamTypes);
+				if(mi == null)
+				{
+					throw new InvalidOperationException();
+				}
+				ilgen.Emit(OpCodes.Call, mi);
+			}
+			else
+			{
+				TypeWrapper tw = ClassLoaderWrapper.LoadClassCritical(Class);
+				MethodWrapper mw = tw.GetMethodWrapper(Name, Sig, false);
+				if(mw == null)
+				{
+					throw new InvalidOperationException();
+				}
+				mw.Link();
+				mw.EmitCall(ilgen);
+			}
+			// TODO we may need a cast here (or a stack to return type conversion)
+			ilgen.Emit(OpCodes.Ret);
+		}
 	}
 
 	public class Override
@@ -1124,6 +1262,8 @@ namespace IKVM.Internal.MapXml
 		public MethodAttributes MethodAttributes;
 		[XmlAttribute("type")]
 		public string Type;
+		[XmlAttribute("nonullcheck")]
+		public bool NoNullCheck;
 		[XmlElement("parameter")]
 		public Param[] Params;
 		public InstructionList body;
@@ -1135,6 +1275,18 @@ namespace IKVM.Internal.MapXml
 		public Throws[] throws;
 		[XmlElement("attribute")]
 		public Attribute[] Attributes;
+
+		internal void Emit(ILGenerator ilgen)
+		{
+			if(redirect != null)
+			{
+				redirect.Emit(ilgen);
+			}
+			else
+			{
+				body.Emit(ilgen);
+			}
+		}
 	}
 
 	public class Field
@@ -1168,6 +1320,8 @@ namespace IKVM.Internal.MapXml
 	{
 		[XmlAttribute("class")]
 		public string Name;
+		[XmlElement("method")]
+		public Method[] Methods;
 	}
 
 	[Flags]
