@@ -24,6 +24,7 @@
 using System;
 using System.Reflection;
 using System.Diagnostics;
+using System.Threading;
 using IKVM.Attributes;
 using IKVM.Internal;
 
@@ -206,7 +207,7 @@ namespace IKVM.Runtime
 			// (to be compatible with Sun)
 			if(obj != null && !DynamicInstanceOf(obj, type, clazz))
 			{
-				throw JavaException.ClassCastException(ClassLoaderWrapper.GetWrapperFromType(obj.GetType()).Name);
+				throw JavaException.ClassCastException(NativeCode.ikvm.runtime.Util.GetTypeWrapperFromObject(obj).Name);
 			}
 			return obj;
 		}
@@ -222,7 +223,7 @@ namespace IKVM.Runtime
 				return false;
 			}
 			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			TypeWrapper other = ClassLoaderWrapper.GetWrapperFromType(obj.GetType());
+			TypeWrapper other = NativeCode.ikvm.runtime.Util.GetTypeWrapperFromObject(obj);
 			return other.IsAssignableTo(wrapper);
 		}
 
@@ -368,7 +369,7 @@ namespace IKVM.Runtime
 		[DebuggerStepThroughAttribute]
 		public static object GetClassFromTypeHandle(RuntimeTypeHandle typeHandle)
 		{
-			return Util.GetClassFromTypeHandle(typeHandle);
+			return NativeCode.ikvm.runtime.Util.getClassFromTypeHandle(typeHandle);
 		}
 
 		[DebuggerStepThroughAttribute]
@@ -476,7 +477,7 @@ namespace IKVM.Runtime
 			{
 				checked
 				{
-					Buffer.BlockCopy(src, srcStart << 3, dest, destStart << 3, len << 3);
+					Buffer.BlockCopy(src, srcStart * 8, dest, destStart * 8, len * 8);
 					return;
 				}
 			}
@@ -501,7 +502,7 @@ namespace IKVM.Runtime
 			{
 				checked
 				{
-					Buffer.BlockCopy(src, srcStart << 2, dest, destStart << 2, len << 2);
+					Buffer.BlockCopy(src, srcStart * 4, dest, destStart * 4, len * 4);
 					return;
 				}
 			}
@@ -526,7 +527,7 @@ namespace IKVM.Runtime
 			{
 				checked
 				{
-					Buffer.BlockCopy(src, srcStart << 1, dest, destStart << 1, len << 1);
+					Buffer.BlockCopy(src, srcStart * 2, dest, destStart * 2, len * 2);
 					return;
 				}
 			}
@@ -579,6 +580,58 @@ namespace IKVM.Runtime
 			return false;
 #else
 			return Environment.HasShutdownStarted && !IKVM.Internal.JVM.Library.runFinalizersOnExit();
+#endif
+		}
+
+#if !WHIDBEY || COMPACT_FRAMEWORK
+		private static readonly object volatileLock = new object();
+#endif
+
+		public static long VolatileRead(ref long v)
+		{
+#if WHIDBEY && !COMPACT_FRAMEWORK
+			return Interlocked.Read(ref v);
+#else
+			lock(volatileLock)
+			{
+				return v;
+			}
+#endif
+		}
+
+		public static void VolatileWrite(ref long v, long newValue)
+		{
+#if WHIDBEY && !COMPACT_FRAMEWORK
+			Interlocked.Exchange(ref v, newValue);
+#else
+			lock(volatileLock)
+			{
+				v = newValue;
+			}
+#endif
+		}
+
+		public static double VolatileRead(ref double v)
+		{
+#if WHIDBEY && !COMPACT_FRAMEWORK
+			return Interlocked.CompareExchange(ref v, 0.0, 0.0);
+#else
+			lock(volatileLock)
+			{
+				return v;
+			}
+#endif
+		}
+
+		public static void VolatileWrite(ref double v, double newValue)
+		{
+#if WHIDBEY && !COMPACT_FRAMEWORK
+			Interlocked.Exchange(ref v, newValue);
+#else
+			lock(volatileLock)
+			{
+				v = newValue;
+			}
 #endif
 		}
 	}
