@@ -21,7 +21,6 @@
   jeroen@frijters.net
   
 */
-#if !COMPACT_FRAMEWORK
 using System;
 using System.IO;
 using System.Collections;
@@ -42,6 +41,30 @@ namespace IKVM.Internal
 		LinkageError
 	}
 
+	[Flags]
+	enum ClassFileParseOptions
+	{
+		None = 0,
+		LocalVariableTable = 1,
+		LineNumberTable = 2,
+	}
+
+	sealed class StringConstants
+	{
+		private StringConstants() {}
+
+		internal static readonly string CLINIT = string.Intern("<clinit>");
+		internal static readonly string INIT = string.Intern("<init>");
+		internal static readonly string SIG_VOID = string.Intern("()V");
+
+		internal static readonly string JAVA_LANG_SYSTEM = string.Intern("java.lang.System");
+		internal static readonly string JAVA_LANG_VMSYSTEM = string.Intern("java.lang.VMSystem");
+		internal static readonly string ARRAYCOPY = string.Intern("arraycopy");
+		internal static readonly string SIG_ARRAYCOPY = string.Intern("(Ljava.lang.Object;ILjava.lang.Object;II)V");
+		internal static readonly string FINALIZE = string.Intern("finalize");
+		internal static readonly string CLONE = string.Intern("clone");
+	}
+
 	sealed class ClassFile
 	{
 		private ConstantPoolItem[] constantpool;
@@ -54,6 +77,7 @@ namespace IKVM.Internal
 		private const ushort FLAG_MASK_MAJORVERSION = 0xFF;
 		private const ushort FLAG_MASK_DEPRECATED = 0x100;
 		private const ushort FLAG_MASK_INTERNAL = 0x200;
+		private const ushort FLAG_MASK_EFFECTIVELY_FINAL = 0x400;
 		private ConstantPoolItemClass[] interfaces;
 		private Field[] fields;
 		private Method[] methods;
@@ -67,27 +91,29 @@ namespace IKVM.Internal
 		private class SupportedVersions
 		{
 			internal static readonly int Minimum = 45;
-			internal static readonly int Maximum = 49;
+			internal static readonly int Maximum = 50;
 		}
 
+#if STATIC_COMPILER
 		// This method parses just enough of the class file to obtain its name, it doesn't
 		// validate the class file structure, but it may throw a ClassFormatError if it
 		// encounters bogus data
 		internal static string GetClassName(byte[] buf, int offset, int length)
 		{
-			const string inputClassName = "(unknown)";
 			BigEndianBinaryReader br = new BigEndianBinaryReader(buf, offset, length);
 			if(br.ReadUInt32() != 0xCAFEBABE)
 			{
-				throw new ClassFormatError("{0} (Bad magic number)", inputClassName);
+				throw new ClassFormatError("Bad magic number");
 			}
 			int minorVersion = br.ReadUInt16();
 			int majorVersion = br.ReadUInt16();
 			if((majorVersion & FLAG_MASK_MAJORVERSION) != majorVersion
 				|| majorVersion < SupportedVersions.Minimum
-				|| majorVersion > SupportedVersions.Maximum)
+				|| majorVersion > SupportedVersions.Maximum
+				|| (majorVersion == SupportedVersions.Minimum && minorVersion < 3)
+				|| (majorVersion == SupportedVersions.Maximum && minorVersion != 0))
 			{
-				throw new UnsupportedClassVersionError(inputClassName + " (" + majorVersion + "." + minorVersion + ")");
+				throw new UnsupportedClassVersionError(majorVersion + "." + minorVersion);
 			}
 			int constantpoolcount = br.ReadUInt16();
 			int[] cpclass = new int[constantpoolcount];
@@ -117,10 +143,10 @@ namespace IKVM.Internal
 						br.Skip(2);
 						break;
 					case Constant.Utf8:
-						utf8_cp[i] = br.ReadString(inputClassName);
+						utf8_cp[i] = br.ReadString("<unknown>");
 						break;
 					default:
-						throw new ClassFormatError("{0} (Illegal constant pool type 0x{1:X})", inputClassName, tag);
+						throw new ClassFormatError("Illegal constant pool type 0x{0:X}", tag);
 				}
 			}
 			br.ReadUInt16(); // access_flags
@@ -133,8 +159,9 @@ namespace IKVM.Internal
 				throw new ClassFormatError("{0}: {1}", x.GetType().Name, x.Message);
 			}
 		}
+#endif // STATIC_COMPILER
 
-		internal ClassFile(byte[] buf, int offset, int length, string inputClassName)
+		internal ClassFile(byte[] buf, int offset, int length, string inputClassName, ClassFileParseOptions options)
 		{
 			try
 			{
@@ -147,7 +174,9 @@ namespace IKVM.Internal
 				ushort majorVersion = br.ReadUInt16();
 				if((majorVersion & FLAG_MASK_MAJORVERSION) != majorVersion
 					|| majorVersion < SupportedVersions.Minimum
-					|| majorVersion > SupportedVersions.Maximum)
+					|| majorVersion > SupportedVersions.Maximum
+					|| (majorVersion == SupportedVersions.Minimum && minorVersion < 3)
+					|| (majorVersion == SupportedVersions.Maximum && minorVersion != 0))
 				{
 					throw new UnsupportedClassVersionError(inputClassName + " (" + majorVersion + "." + minorVersion + ")");
 				}
@@ -279,7 +308,7 @@ namespace IKVM.Internal
 				{
 					fields[i] = new Field(this, br);
 					string name = fields[i].Name;
-					if(!IsValidFieldName(name))
+					if(!IsValidFieldName(name, majorVersion))
 					{
 						throw new ClassFormatError("{0} (Illegal field name \"{1}\")", Name, name);
 					}
@@ -295,16 +324,19 @@ namespace IKVM.Internal
 				methods = new Method[methods_count];
 				for(int i = 0; i < methods_count; i++)
 				{
-					methods[i] = new Method(this, br);
+					methods[i] = new Method(this, options, br);
 					string name = methods[i].Name;
 					string sig = methods[i].Signature;
-					if(!IsValidMethodName(name) && !ReferenceEquals(name, "<init>") && !ReferenceEquals(name, "<clinit>"))
+					if(!IsValidMethodName(name, majorVersion))
 					{
-						throw new ClassFormatError("{0} (Illegal method name \"{1}\")", Name, name);
-					}
-					if((ReferenceEquals(name, "<init>") || ReferenceEquals(name, "<clinit>")) && !sig.EndsWith("V"))
-					{
-						throw new ClassFormatError("{0} (Method \"{1}\" has illegal signature \"{2}\")", Name, name, sig);
+						if(!ReferenceEquals(name, StringConstants.INIT) && !ReferenceEquals(name, StringConstants.CLINIT))
+						{
+							throw new ClassFormatError("{0} (Illegal method name \"{1}\")", Name, name);
+						}
+						if(!sig.EndsWith("V"))
+						{
+							throw new ClassFormatError("{0} (Method \"{1}\" has illegal signature \"{2}\")", Name, name, sig);
+						}
 					}
 					for(int j = 0; j < i; j++)
 					{
@@ -320,8 +352,11 @@ namespace IKVM.Internal
 					switch(GetConstantPoolUtf8String(br.ReadUInt16()))
 					{
 						case "Deprecated":
+							if(br.ReadUInt32() != 0)
+							{
+								throw new ClassFormatError("Invalid Deprecated attribute length");
+							}
 							flags |= FLAG_MASK_DEPRECATED;
-							br.Skip(br.ReadUInt32());
 							break;
 						case "SourceFile":
 							if(br.ReadUInt32() != 2)
@@ -332,12 +367,13 @@ namespace IKVM.Internal
 							break;
 						case "InnerClasses":
 						{
-							// Sun totally ignores the length of InnerClasses attribute,
-							// so when we're running Fuzz this used to show up as lots of differences,
-							// now we do the same.
 							BigEndianBinaryReader rdr = br;
-							br.ReadUInt32();
+							uint attribute_length = br.ReadUInt32();
 							ushort count = rdr.ReadUInt16();
+							if(this.MajorVersion >= 49 && attribute_length != 2 + count * (2 + 2 + 2 + 2))
+							{
+								throw new ClassFormatError("{0} (InnerClasses attribute has incorrect length)", this.Name);
+							}
 							innerClasses = new InnerClass[count];
 							for(int j = 0; j < innerClasses.Length; j++)
 							{
@@ -564,7 +600,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		private static bool IsValidMethodName(string name)
+		private static bool IsValidMethodName(string name, int majorVersion)
 		{
 			if(name.Length == 0)
 			{
@@ -572,15 +608,15 @@ namespace IKVM.Internal
 			}
 			for(int i = 0; i < name.Length; i++)
 			{
-				if(".;[/<>".IndexOf(name[i]) != -1)
+				if(".;/<>".IndexOf(name[i]) != -1)
 				{
 					return false;
 				}
 			}
-			return true;
+			return majorVersion >= 49 || IsValidPre49Identifier(name);
 		}
 
-		private static bool IsValidFieldName(string name)
+		private static bool IsValidFieldName(string name, int majorVersion)
 		{
 			if(name.Length == 0)
 			{
@@ -588,7 +624,23 @@ namespace IKVM.Internal
 			}
 			for(int i = 0; i < name.Length; i++)
 			{
-				if(".;[/".IndexOf(name[i]) != -1)
+				if(".;/".IndexOf(name[i]) != -1)
+				{
+					return false;
+				}
+			}
+			return majorVersion >= 49 || IsValidPre49Identifier(name);
+		}
+
+		private static bool IsValidPre49Identifier(string name)
+		{
+			if(!Char.IsLetter(name[0]) && "$_".IndexOf(name[0]) == -1)
+			{
+				return false;
+			}
+			for(int i = 1; i < name.Length; i++)
+			{
+				if(!Char.IsLetterOrDigit(name[i]) && "$_".IndexOf(name[i]) == -1)
 				{
 					return false;
 				}
@@ -610,7 +662,6 @@ namespace IKVM.Internal
 			switch(sig[start])
 			{
 				case 'L':
-					// TODO we might consider doing more checking here
 					return sig.IndexOf(';', start + 1) == end - 1;
 				case '[':
 					while(sig[start] == '[')
@@ -815,6 +866,17 @@ namespace IKVM.Internal
 			return (ConstantPoolItemFieldref)constantpool[index];
 		}
 
+		// this won't throw an exception if index is invalid
+		// (used by IsSideEffectFreeStaticInitializer)
+		internal ConstantPoolItemFieldref SafeGetFieldref(int index)
+		{
+			if(index > 0 && index < constantpool.Length)
+			{
+				return constantpool[index] as ConstantPoolItemFieldref;
+			}
+			return null;
+		}
+
 		// NOTE this returns an MI, because it used for both normal methods and interface methods
 		internal ConstantPoolItemMI GetMethodref(int index)
 		{
@@ -831,6 +893,17 @@ namespace IKVM.Internal
 			return ((ConstantPoolItemClass)constantpool[index]).Name;
 		}
 
+		// this won't throw an exception if index is invalid
+		// (used by IsSideEffectFreeStaticInitializer)
+		internal bool SafeIsConstantPoolClass(int index)
+		{
+			if(index > 0 && index < constantpool.Length)
+			{
+				return constantpool[index] as ConstantPoolItemClass != null;
+			}
+			return false;
+		}
+
 		internal TypeWrapper GetConstantPoolClassType(int index)
 		{
 			return ((ConstantPoolItemClass)constantpool[index]).GetClassType();
@@ -986,6 +1059,19 @@ namespace IKVM.Internal
 			flags |= FLAG_MASK_INTERNAL;
 		}
 
+		internal void SetEffectivelyFinal()
+		{
+			flags |= FLAG_MASK_EFFECTIVELY_FINAL;
+		}
+
+		internal bool IsEffectivelyFinal
+		{
+			get
+			{
+				return (flags & FLAG_MASK_EFFECTIVELY_FINAL) != 0;
+			}
+		}
+
 		internal struct InnerClass
 		{
 			internal ushort innerClass;		// ConstantPoolItemClass
@@ -1033,6 +1119,7 @@ namespace IKVM.Internal
 			private ushort name_index;
 			private string name;
 			private TypeWrapper typeWrapper;
+			private static char[] invalidJava15Characters = { '.', ';' };
 
 			internal ConstantPoolItemClass(BigEndianBinaryReader br)
 			{
@@ -1044,35 +1131,63 @@ namespace IKVM.Internal
 				name = classFile.GetConstantPoolUtf8String(name_index);
 				if(name.Length > 0)
 				{
-					char prev = name[0];
-					if(Char.IsLetter(prev) || prev == '$' || prev == '_' || prev == '[')
+					if(classFile.MajorVersion < 49)
 					{
-						int skip = 1;
+						char prev = name[0];
+						if(Char.IsLetter(prev) || prev == '$' || prev == '_' || prev == '[' || prev == '/')
+						{
+							int skip = 1;
+							int end = name.Length;
+							if(prev == '[')
+							{
+								if(!IsValidFieldSig(name))
+								{
+									goto barf;
+								}
+								while(name[skip] == '[')
+								{
+									skip++;
+								}
+								if(name.EndsWith(";"))
+								{
+									end--;
+								}
+							}
+							for(int i = skip; i < end; i++)
+							{
+								char c = name[i];
+								if(!Char.IsLetterOrDigit(c) && c != '$' && c != '_' && (c != '/' || prev == '/'))
+								{
+									goto barf;
+								}
+								prev = c;
+							}
+							name = String.Intern(name.Replace('/', '.'));
+							return;
+						}
+					}
+					else
+					{
+						// since 1.5 the restrictions on class names have been greatly reduced
 						int end = name.Length;
-						if(prev == '[')
+						if(name[0] == '[')
 						{
-							// TODO optimize this
 							if(!IsValidFieldSig(name))
 							{
 								goto barf;
 							}
-							while(name[skip] == '[')
-							{
-								skip++;
-							}
-							if(name.EndsWith(";"))
+							// the semicolon is only allowed at the end and IsValidFieldSig enforces this,
+							// but since invalidJava15Characters contains the semicolon, we decrement end
+							// to make the following check against invalidJava15Characters ignore the
+							// trailing semicolon.
+							if(name[end - 1] == ';')
 							{
 								end--;
 							}
 						}
-						for(int i = skip; i < end; i++)
+						if(name.IndexOfAny(invalidJava15Characters, 0, end) >= 0)
 						{
-							char c = name[i];
-							if(!Char.IsLetterOrDigit(c) && c != '$' && c != '_' && c != '-' && (c != '/' || prev == '/'))
-							{
-								goto barf;
-							}
-							prev = c;
+							goto barf;
 						}
 						name = String.Intern(name.Replace('/', '.'));
 						return;
@@ -1130,7 +1245,7 @@ namespace IKVM.Internal
 			{
 				// HACK keep the compiler from warning about unused local
 				GC.KeepAlive(x);
-#if !STATIC_COMPILER
+#if !STATIC_COMPILER && !COMPACT_FRAMEWORK
 				if(Tracer.ClassLoading.TraceError)
 				{
 					object cl = classLoader.GetJavaClassLoader();
@@ -1155,7 +1270,7 @@ namespace IKVM.Internal
 						}
 						Tracer.Error(Tracer.ClassLoading, "ClassLoader chain: {0}", sb);
 					}
-					Exception m = IKVM.Runtime.Util.MapException(x.ToJava());
+					Exception m = JVM.Library.mapException(x.ToJava());
 					Tracer.Error(Tracer.ClassLoading, m.ToString() + Environment.NewLine + m.StackTrace);
 				}
 #endif // !STATIC_COMPILER
@@ -1304,11 +1419,11 @@ namespace IKVM.Internal
 				}
 				name = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.name_index));
 				descriptor = classFile.GetConstantPoolUtf8String(name_and_type.descriptor_index);
-				Validate(name, descriptor);
+				Validate(name, descriptor, classFile.MajorVersion);
 				descriptor = String.Intern(descriptor.Replace('/', '.'));
 			}
 
-			protected abstract void Validate(string name, string descriptor);
+			protected abstract void Validate(string name, string descriptor, int majorVersion);
 
 			internal override void Link(TypeWrapper thisType, Hashtable classCache)
 			{
@@ -1354,13 +1469,13 @@ namespace IKVM.Internal
 			{
 			}
 
-			protected override void Validate(string name, string descriptor)
+			protected override void Validate(string name, string descriptor, int majorVersion)
 			{
 				if(!IsValidFieldSig(descriptor))
 				{
 					throw new ClassFormatError("Invalid field signature \"{0}\"", descriptor);
 				}
-				if(!IsValidFieldName(name))
+				if(!IsValidFieldName(name, majorVersion))
 				{
 					throw new ClassFormatError("Invalid field name \"{0}\"", name);
 				}
@@ -1420,23 +1535,23 @@ namespace IKVM.Internal
 			{
 			}
 
-			protected override void Validate(string name, string descriptor)
+			protected override void Validate(string name, string descriptor, int majorVersion)
 			{
 				if(!IsValidMethodSig(descriptor))
 				{
 					throw new ClassFormatError("Method {0} has invalid signature {1}", name, descriptor);
 				}
-				if(ReferenceEquals(name, "<init>") || ReferenceEquals(name, "<clinit>"))
+				if(!IsValidMethodName(name, majorVersion))
 				{
+					if(!ReferenceEquals(name, StringConstants.INIT))
+					{
+						throw new ClassFormatError("Invalid method name \"{0}\"", name);
+					}
 					if(!descriptor.EndsWith("V"))
 					{
 						throw new ClassFormatError("Method {0} has invalid signature {1}", name, descriptor);
 					}
 				}
-				else if(!IsValidMethodName(name))
-				{
-					throw new ClassFormatError("Invalid method name \"{0}\"", name);
-				}
 			}
 
 			internal override void Link(TypeWrapper thisType, Hashtable classCache)
@@ -1495,12 +1610,12 @@ namespace IKVM.Internal
 				TypeWrapper wrapper = GetClassType();
 				if(!wrapper.IsUnloadable)
 				{
-					method = wrapper.GetMethodWrapper(Name, Signature, !ReferenceEquals(Name, "<init>"));
+					method = wrapper.GetMethodWrapper(Name, Signature, !ReferenceEquals(Name, StringConstants.INIT));
 					if(method != null)
 					{
 						method.Link();
 					}
-					if(Name != "<init>" && 
+					if(Name != StringConstants.INIT && 
 						(thisType.Modifiers & (Modifiers.Interface | Modifiers.Super)) == Modifiers.Super &&
 						thisType != wrapper && thisType.IsSubTypeOf(wrapper))
 					{
@@ -1628,7 +1743,7 @@ namespace IKVM.Internal
 			}
 		}
 
-		internal sealed class ConstantPoolItemNameAndType : ConstantPoolItem
+		private sealed class ConstantPoolItemNameAndType : ConstantPoolItem
 		{
 			internal ushort name_index;
 			internal ushort descriptor_index;
@@ -1638,6 +1753,15 @@ namespace IKVM.Internal
 				name_index = br.ReadUInt16();
 				descriptor_index = br.ReadUInt16();
 			}
+
+			internal override void Resolve(ClassFile classFile)
+			{
+				if(classFile.GetConstantPoolUtf8String(name_index) == null
+					|| classFile.GetConstantPoolUtf8String(descriptor_index) == null)
+				{
+					throw new ClassFormatError("Illegal constant pool index");
+				}
+			}
 		}
 
 		private sealed class ConstantPoolItemString : ConstantPoolItem
@@ -1868,8 +1992,11 @@ namespace IKVM.Internal
 					switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
 					{
 						case "Deprecated":
+							if(br.ReadUInt32() != 0)
+							{
+								throw new ClassFormatError("Invalid Deprecated attribute length");
+							}
 							flags |= FLAG_MASK_DEPRECATED;
-							br.Skip(br.ReadUInt32());
 							break;
 						case "ConstantValue":
 						{
@@ -1996,10 +2123,10 @@ namespace IKVM.Internal
 			private object annotationDefault;
 			private object[][] parameterAnnotations;
 
-			internal Method(ClassFile classFile, BigEndianBinaryReader br) : base(classFile, br)
+			internal Method(ClassFile classFile, ClassFileParseOptions options, BigEndianBinaryReader br) : base(classFile, br)
 			{
 				// vmspec 4.6 says that all flags, except ACC_STRICT are ignored on <clinit>
-				if(ReferenceEquals(Name, "<clinit>") && ReferenceEquals(Signature, "()V"))
+				if(ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID))
 				{
 					access_flags &= Modifiers.Strictfp;
 					access_flags |= (Modifiers.Static | Modifiers.Private);
@@ -2008,7 +2135,7 @@ namespace IKVM.Internal
 				{
 					// LAMESPEC: vmspec 4.6 says that abstract methods can not be strictfp (and this makes sense), but
 					// javac (pre 1.5) is broken and marks abstract methods as strictfp (if you put the strictfp on the class)
-					if((ReferenceEquals(Name, "<init>") && (IsStatic || IsSynchronized || IsFinal || IsAbstract || IsNative))
+					if((ReferenceEquals(Name, StringConstants.INIT) && (IsStatic || IsSynchronized || IsFinal || IsAbstract || IsNative))
 						|| (IsPrivate && IsPublic) || (IsPrivate && IsProtected) || (IsPublic && IsProtected)
 						|| (IsAbstract && (IsFinal || IsNative || IsPrivate || IsStatic || IsSynchronized))
 						|| (classFile.IsInterface && (!IsPublic || !IsAbstract)))
@@ -2022,8 +2149,11 @@ namespace IKVM.Internal
 					switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
 					{
 						case "Deprecated":
+							if(br.ReadUInt32() != 0)
+							{
+								throw new ClassFormatError("Invalid Deprecated attribute length");
+							}
 							flags |= FLAG_MASK_DEPRECATED;
-							br.Skip(br.ReadUInt32());
 							break;
 						case "Code":
 						{
@@ -2032,7 +2162,7 @@ namespace IKVM.Internal
 								throw new ClassFormatError("{0} (Duplicate Code attribute)", classFile.Name);
 							}
 							BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
-							code.Read(classFile, this, rdr);
+							code.Read(classFile, this, rdr, options);
 							if(!rdr.IsAtEnd)
 							{
 								throw new ClassFormatError("{0} (Code attribute has wrong length)", classFile.Name);
@@ -2146,7 +2276,7 @@ namespace IKVM.Internal
 				{
 					if(code.IsEmpty)
 					{
-						if(ReferenceEquals(this.Name, "<clinit>"))
+						if(ReferenceEquals(this.Name, StringConstants.CLINIT))
 						{
 							code.verifyError = string.Format("Class {0}, method {1} signature {2}: No Code attribute", classFile.Name, this.Name, this.Signature);
 							return;
@@ -2177,7 +2307,7 @@ namespace IKVM.Internal
 			{
 				get
 				{
-					return ReferenceEquals(Name, "<clinit>") && ReferenceEquals(Signature, "()V");
+					return ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID);
 				}
 			}
 
@@ -2291,12 +2421,12 @@ namespace IKVM.Internal
 				internal LineNumberTableEntry[] lineNumberTable;
 				internal LocalVariableTableEntry[] localVariableTable;
 
-				internal void Read(ClassFile classFile, Method method, BigEndianBinaryReader br)
+				internal void Read(ClassFile classFile, Method method, BigEndianBinaryReader br, ClassFileParseOptions options)
 				{
 					max_stack = br.ReadUInt16();
 					max_locals = br.ReadUInt16();
 					uint code_length = br.ReadUInt32();
-					if(code_length > 65536)
+					if(code_length > 65535)
 					{
 						throw new ClassFormatError("{0} (Invalid Code length {1})", classFile.Name, code_length);
 					}
@@ -2324,11 +2454,22 @@ namespace IKVM.Internal
 					exception_table = new ExceptionTableEntry[exception_table_length];
 					for(int i = 0; i < exception_table_length; i++)
 					{
+						ushort start_pc = br.ReadUInt16();
+						ushort end_pc = br.ReadUInt16();
+						ushort handler_pc = br.ReadUInt16();
+						ushort catch_type = br.ReadUInt16();
+						if(start_pc >= end_pc
+							|| end_pc > code_length
+							|| handler_pc >= code_length
+							|| (catch_type != 0 && !classFile.SafeIsConstantPoolClass(catch_type)))
+						{
+							throw new ClassFormatError("Illegal exception table: {0}.{1}{2}", classFile.Name, method.Name, method.Signature);
+						}
 						exception_table[i] = new ExceptionTableEntry();
-						exception_table[i].start_pc = br.ReadUInt16();
-						exception_table[i].end_pc = br.ReadUInt16();
-						exception_table[i].handler_pc = br.ReadUInt16();
-						exception_table[i].catch_type = br.ReadUInt16();
+						exception_table[i].start_pc = start_pc;
+						exception_table[i].end_pc = end_pc;
+						exception_table[i].handler_pc = handler_pc;
+						exception_table[i].catch_type = catch_type;
 						exception_table[i].ordinal = i;
 					}
 					ushort attributes_count = br.ReadUInt16();
@@ -2337,11 +2478,7 @@ namespace IKVM.Internal
 						switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
 						{
 							case "LineNumberTable":
-								if(JVM.NoStackTraceInfo)
-								{
-									br.Skip(br.ReadUInt32());
-								}
-								else
+								if((options & ClassFileParseOptions.LineNumberTable) != 0)
 								{
 									BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
 									int count = rdr.ReadUInt16();
@@ -2360,9 +2497,13 @@ namespace IKVM.Internal
 										throw new ClassFormatError("{0} (LineNumberTable attribute has wrong length)", classFile.Name);
 									}
 								}
+								else
+								{
+									br.Skip(br.ReadUInt32());
+								}
 								break;
 							case "LocalVariableTable":
-								if(JVM.Debug || JVM.IsStaticCompiler)
+								if((options & ClassFileParseOptions.LocalVariableTable) != 0)
 								{
 									BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
 									int count = rdr.ReadUInt16();
@@ -2728,4 +2869,3 @@ namespace IKVM.Internal
 		}
 	}
 }
-#endif
