@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002, 2003, 2004, 2005, 2006 Jeroen Frijters
+  Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -40,7 +40,6 @@ namespace IKVM.Internal
 		None = 0,
 		HideFromReflection = 1,
 		ExplicitOverride = 2,
-		LiteralField = 4,
 		MirandaMethod = 8,
 		AccessStub = 16,
 		InternalAccess = 32  // member has "internal" access (@ikvm.lang.Internal)
@@ -136,7 +135,7 @@ namespace IKVM.Internal
 				return IsPublic ||
 					caller == DeclaringType ||
 					(IsProtected && caller.IsSubTypeOf(DeclaringType) && (IsStatic || instance.IsSubTypeOf(caller))) ||
-					(IsInternal && caller.Assembly == DeclaringType.Assembly) ||
+					(IsInternal && caller.GetClassLoader() == DeclaringType.GetClassLoader()) ||
 					(!IsPrivate && caller.IsInSamePackageAs(DeclaringType));
 			}
 			return false;
@@ -158,19 +157,6 @@ namespace IKVM.Internal
 			}
 		}
 
-		internal bool IsLiteralField
-		{
-			get
-			{
-				return (flags & MemberFlags.LiteralField) != 0;
-			}
-			set
-			{
-				flags &= ~MemberFlags.LiteralField;
-				flags |= value ? MemberFlags.LiteralField : MemberFlags.None;
-			}
-		}
-
 		internal bool IsMirandaMethod
 		{
 			get
@@ -263,7 +249,12 @@ namespace IKVM.Internal
 			throw new InvalidOperationException();
 		}
 
-		internal virtual void EmitNewobj(ILGenerator ilgen)
+		internal void EmitNewobj(ILGenerator ilgen)
+		{
+			EmitNewobj(ilgen, null, 0);
+		}
+
+		internal virtual void EmitNewobj(ILGenerator ilgen, MethodAnalyzer ma, int opcodeIndex)
 		{
 			throw new InvalidOperationException();
 		}
@@ -369,7 +360,7 @@ namespace IKVM.Internal
 		{
 			get
 			{
-				return method != null;
+				return parameterTypeWrappers != null;
 			}
 		}
 
@@ -467,12 +458,16 @@ namespace IKVM.Internal
 			// NOTE if method is a MethodBuilder, GetCustomAttributes doesn't work (and if
 			// the method had any declared exceptions, the declaredExceptions field would have
 			// been set)
+#if COMPACT_FRAMEWORK
+			if(method != null)
+#else
 			if(method != null && !(method is MethodBuilder))
+#endif
 			{
-				object[] attributes = method.GetCustomAttributes(typeof(ThrowsAttribute), false);
-				if(attributes.Length == 1)
+				ThrowsAttribute attr = AttributeHelper.GetThrows(method);
+				if(attr != null)
 				{
-					return ((ThrowsAttribute)attributes[0]).Classes;
+					return attr.Classes;
 				}
 			}
 			return new string[0];
@@ -550,6 +545,7 @@ namespace IKVM.Internal
 		internal virtual object Invoke(object obj, object[] args, bool nonVirtual)
 		{
 			AssertLinked();
+#if !COMPACT_FRAMEWORK
 			// if we've still got the builder object, we need to replace it with the real thing before we can call it
 			if(method is MethodBuilder)
 			{
@@ -589,6 +585,7 @@ namespace IKVM.Internal
 					throw new InvalidOperationException("Failed to fixate constructor: " + this.DeclaringType.Name + "." + this.Name + this.Signature);
 				}
 			}
+#endif // !COMPACT_FRAMEWORK
 			return InvokeImpl(method, obj, args, nonVirtual);
 		}
 
@@ -597,7 +594,9 @@ namespace IKVM.Internal
 		[HideFromJava]
 		internal object InvokeImpl(MethodBase method, object obj, object[] args, bool nonVirtual)
 		{
+#if !COMPACT_FRAMEWORK
 			Debug.Assert(!(method is MethodBuilder || method is ConstructorBuilder));
+#endif // !COMPACT_FRAMEWORK
 
 			if(IsStatic)
 			{
@@ -606,7 +605,7 @@ namespace IKVM.Internal
 			}
 			else
 			{
-				if(ReferenceEquals(Name, "<init>"))
+				if(ReferenceEquals(Name, StringConstants.INIT))
 				{
 					if(method is MethodInfo)
 					{
@@ -638,7 +637,7 @@ namespace IKVM.Internal
 						}
 						catch(TargetInvocationException x)
 						{
-							throw JavaException.InvocationTargetException(IKVM.Runtime.Util.MapException(x.InnerException));
+							throw JavaException.InvocationTargetException(JVM.Library.mapException(x.InnerException));
 						}
 					}
 					else if(!method.DeclaringType.IsInstanceOfType(obj))
@@ -682,7 +681,7 @@ namespace IKVM.Internal
 					}
 					catch(TargetInvocationException x)
 					{
-						throw JavaException.InvocationTargetException(IKVM.Runtime.Util.MapException(x.InnerException));
+						throw JavaException.InvocationTargetException(JVM.Library.mapException(x.InnerException));
 					}
 #endif
 				}
@@ -704,7 +703,7 @@ namespace IKVM.Internal
 			}
 			catch(TargetInvocationException x)
 			{
-				throw JavaException.InvocationTargetException(IKVM.Runtime.Util.MapException(x.InnerException));
+				throw JavaException.InvocationTargetException(JVM.Library.mapException(x.InnerException));
 			}
 		}
 
@@ -752,8 +751,8 @@ namespace IKVM.Internal
 				cache = new Hashtable(keygen, keygen);
 				AssemblyName name = new AssemblyName();
 				name.Name = "NonvirtualInvoker";
-				AssemblyBuilder ab = AppDomain.CurrentDomain.DefineDynamicAssembly(name, DynamicClassLoader.IsSaveDebugImage ? AssemblyBuilderAccess.RunAndSave : AssemblyBuilderAccess.Run);
-				if(DynamicClassLoader.IsSaveDebugImage)
+				AssemblyBuilder ab = AppDomain.CurrentDomain.DefineDynamicAssembly(name, JVM.IsSaveDebugImage ? AssemblyBuilderAccess.RunAndSave : AssemblyBuilderAccess.Run);
+				if(JVM.IsSaveDebugImage)
 				{
 					module = ab.DefineDynamicModule("NonvirtualInvoker", "NonvirtualInvoker.dll");
 					DynamicClassLoader.RegisterForSaveDebug(ab);
@@ -956,7 +955,7 @@ namespace IKVM.Internal
 			throw new InvalidOperationException();
 		}
 
-		internal sealed override void EmitNewobj(ILGenerator ilgen)
+		internal sealed override void EmitNewobj(ILGenerator ilgen, MethodAnalyzer ma, int opcodeIndex)
 		{
 			AssertLinked();
 			PreEmit(ilgen);
@@ -1073,6 +1072,7 @@ namespace IKVM.Internal
 
 		internal static void IssueWarning(FieldInfo field)
 		{
+#if !COMPACT_FRAMEWORK
 			// FXBUG .NET (1.0 & 1.1)
 			// FieldInfo.GetValue() on a literal causes the type initializer to run and
 			// we don't want that.
@@ -1092,6 +1092,7 @@ namespace IKVM.Internal
 					}
 				}
 			}
+#endif // !COMPACT_FRAMEWORK
 		}
 
 		private class Helper
@@ -1122,7 +1123,7 @@ namespace IKVM.Internal
 
 		internal FieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, string name, string sig, ExModifiers modifiers, FieldInfo field)
 			: this(declaringType, fieldType, name, sig, modifiers.Modifiers, field,
-					(modifiers.IsInternal ? MemberFlags.InternalAccess : MemberFlags.None) | (field != null && field.IsLiteral ? MemberFlags.LiteralField : MemberFlags.None))
+					(modifiers.IsInternal ? MemberFlags.InternalAccess : MemberFlags.None))
 		{
 		}
 
@@ -1159,26 +1160,27 @@ namespace IKVM.Internal
 #if WHIDBEY
 					val = field.GetRawConstantValue();
 #else
-					val = field.GetValue(null);
-					if(val is Enum)
+					try
 					{
-						val = DotNetTypeWrapper.EnumValueFieldWrapper.GetEnumPrimitiveValue(val);
+						val = field.GetValue(null);
+					}
+					catch(TargetInvocationException x)
+					{
+						throw x.InnerException;
 					}
 #endif
+					if(field.FieldType.IsEnum)
+					{
+						val = DotNetTypeWrapper.EnumValueFieldWrapper.GetEnumPrimitiveValue(Enum.GetUnderlyingType(field.FieldType), val);
+					}
 				}
 				else
 				{
-#if WHIDBEY
-					// TODO
-#else
-					// In Java, instance fields can also have a ConstantValue attribute so we emulate that
-					// with ConstantValueAttribute (for consumption by ikvmstub only)
-					object[] attrib = field.GetCustomAttributes(typeof(ConstantValueAttribute), false);
-					if(attrib.Length == 1)
-					{
-						val = ((ConstantValueAttribute)attrib[0]).GetConstantValue();
-					}
-#endif
+					// NOTE instance fields can also be "constant" and we round trip this information to make the Japi results look
+					// nice (but otherwise this has no practical value), but note that this only works when the code is compiled
+					// with -strictfieldfieldsemantics (because the ConstantValueAttribute is on the field and when we're a GetterFieldWrapper
+					// we don't have access to the corresponding field).
+					val = AttributeHelper.GetConstantValue(field);
 				}
 				if(val != null && !(val is string))
 				{
@@ -1272,32 +1274,76 @@ namespace IKVM.Internal
 			return new SimpleFieldWrapper(declaringType, fieldType, fi, name, sig, modifiers);
 		}
 
+		private FieldInfo TokenBasedLookup(BindingFlags bindings, int token)
+		{
+			ModuleBuilder module = (ModuleBuilder)DeclaringType.TypeAsTBD.Module;
+			foreach(FieldInfo f in DeclaringType.TypeAsTBD.GetFields(bindings))
+			{
+				if(module.GetFieldToken(f).Token == token)
+				{
+					return f;
+				}
+			}
+			if(Type.GetType("Mono.Runtime") != null)
+			{
+				// MONOBUG token based lookup doesn't work on Mono 1.1.17,
+				// so we'll try again but now do a name/type based comparison
+				// (note that this is not water tight, because of erased types)
+				foreach(FieldInfo f in DeclaringType.TypeAsTBD.GetFields(bindings))
+				{
+					if(f.Name == field.Name && f.FieldType.Equals(field.FieldType))
+					{
+						return f;
+					}
+				}
+			}
+			throw new InvalidOperationException();
+		}
+
 		private void LookupField()
 		{
-			BindingFlags bindings = BindingFlags.Public | BindingFlags.NonPublic;
-			if(IsStatic)
+			FieldBuilder fb = field as FieldBuilder;
+			if(fb != null)
 			{
-				bindings |= BindingFlags.Static;
+#if WHIDBEY
+				field = DeclaringType.TypeAsTBD.Module.ResolveField(fb.GetToken().Token);
+#else
+				// first do a name based lookup as that is much faster than doing a token based lookup
+				BindingFlags bindings = BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.DeclaredOnly;
+				if(this.IsStatic)
+				{
+					bindings |= BindingFlags.Static;
+				}
+				else
+				{
+					bindings |= BindingFlags.Instance;
+				}
+				try
+				{
+					FieldInfo fi = DeclaringType.TypeAsTBD.GetField(field.Name, bindings);
+					// now check that we've got the right field by comparing the tokens
+					ModuleBuilder module = (ModuleBuilder)DeclaringType.TypeAsTBD.Module;
+					if(module.GetFieldToken(fi).Token != fb.GetToken().Token)
+					{
+						fi = TokenBasedLookup(bindings, fb.GetToken().Token);
+					}
+					field = fi;
+				}
+				catch(AmbiguousMatchException)
+				{
+					// .NET 2.0 will throw this exception if there are multiple fields
+					// with the same name (.NET 1.1 will simply return one of them)
+					field = TokenBasedLookup(bindings, fb.GetToken().Token);
+				}
+#endif
 			}
-			else
-			{
-				bindings |= BindingFlags.Instance;
-			}
-			// TODO instead of looking up the field by name, we should use the Token to find it.
-			field = DeclaringType.TypeAsTBD.GetField(Name, bindings);
-			this.IsLiteralField = field.IsLiteral;
-			Debug.Assert(field != null);
 		}
 
 #if !STATIC_COMPILER
 		internal virtual void SetValue(object obj, object val)
 		{
 			AssertLinked();
-			// TODO this is a broken implementation (for one thing, it needs to support redirection)
-			if(field == null || field is FieldBuilder)
-			{
-				LookupField();
-			}
+			LookupField();
 			if(fieldType.IsGhost)
 			{
 				object temp = field.GetValue(obj);
@@ -1318,22 +1364,10 @@ namespace IKVM.Internal
 		internal virtual object GetValue(object obj)
 		{
 			AssertLinked();
-			// TODO this is a broken implementation (for one thing, it needs to support redirection)
-			if(field == null || field is FieldBuilder)
-			{
-				LookupField();
-			}
+			LookupField();
 #if STATIC_COMPILER
 			return field.GetValue(null);
 #else
-			// FieldInfo.IsLiteral is expensive, so we have our own flag
-			// TODO we might be able to ensure that we always use ConstantFieldWrapper for literal fields,
-			// in that case the we could simply remove the check altogether.
-			if(IsLiteralField)
-			{
-				// on a non-broken CLR GetValue on a literal will not trigger type initialization, but on Java it should
-				DeclaringType.RunClassInit();
-			}
 			object val = field.GetValue(obj);
 			if(fieldType.IsGhost)
 			{
@@ -1387,11 +1421,6 @@ namespace IKVM.Internal
 
 	sealed class VolatileLongDoubleFieldWrapper : FieldWrapper
 	{
-		private static MethodInfo volatileReadDouble = typeof(System.Threading.Thread).GetMethod("VolatileRead", new Type[] { Type.GetType("System.Double&") });
-		private static MethodInfo volatileReadLong = typeof(System.Threading.Thread).GetMethod("VolatileRead", new Type[] { Type.GetType("System.Int64&") });
-		private static MethodInfo volatileWriteDouble = typeof(System.Threading.Thread).GetMethod("VolatileWrite", new Type[] { Type.GetType("System.Double&"), typeof(double) });
-		private static MethodInfo volatileWriteLong = typeof(System.Threading.Thread).GetMethod("VolatileWrite", new Type[] { Type.GetType("System.Int64&"), typeof(long) });
-
 		internal VolatileLongDoubleFieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, FieldInfo fi, string name, string sig, ExModifiers modifiers)
 			: base(declaringType, fieldType, name, sig, modifiers, fi)
 		{
@@ -1417,12 +1446,12 @@ namespace IKVM.Internal
 			}
 			if(FieldTypeWrapper == PrimitiveTypeWrapper.DOUBLE)
 			{
-				ilgen.Emit(OpCodes.Call, volatileReadDouble);
+				ilgen.Emit(OpCodes.Call, ByteCodeHelperMethods.volatileReadDouble);
 			}
 			else
 			{
 				Debug.Assert(FieldTypeWrapper == PrimitiveTypeWrapper.LONG);
-				ilgen.Emit(OpCodes.Call, volatileReadLong);
+				ilgen.Emit(OpCodes.Call, ByteCodeHelperMethods.volatileReadLong);
 			}
 		}
 
@@ -1446,12 +1475,12 @@ namespace IKVM.Internal
 			ilgen.Emit(OpCodes.Ldloc, temp);
 			if(FieldTypeWrapper == PrimitiveTypeWrapper.DOUBLE)
 			{
-				ilgen.Emit(OpCodes.Call, volatileWriteDouble);
+				ilgen.Emit(OpCodes.Call, ByteCodeHelperMethods.volatileWriteDouble);
 			}
 			else
 			{
 				Debug.Assert(FieldTypeWrapper == PrimitiveTypeWrapper.LONG);
-				ilgen.Emit(OpCodes.Call, volatileWriteLong);
+				ilgen.Emit(OpCodes.Call, ByteCodeHelperMethods.volatileWriteLong);
 			}
 		}
 #endif
@@ -1460,14 +1489,16 @@ namespace IKVM.Internal
 	sealed class GetterFieldWrapper : FieldWrapper
 	{
 		private MethodInfo getter;
+		private PropertyInfo prop;
 
 		// NOTE fi may be null!
-		internal GetterFieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, FieldInfo fi, string name, string sig, ExModifiers modifiers, MethodInfo getter)
+		internal GetterFieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, FieldInfo fi, string name, string sig, ExModifiers modifiers, MethodInfo getter, PropertyInfo prop)
 			: base(declaringType, fieldType, name, sig, modifiers, fi)
 		{
 			Debug.Assert(!IsVolatile);
 
 			this.getter = getter;
+			this.prop = prop;
 		}
 
 		internal void SetGetter(MethodInfo getter)
@@ -1475,6 +1506,34 @@ namespace IKVM.Internal
 			this.getter = getter;
 		}
 
+		internal MethodInfo GetGetter()
+		{
+			return getter;
+		}
+
+		internal PropertyInfo GetProperty()
+		{
+			return prop;
+		}
+
+#if !STATIC_COMPILER
+		internal override object GetValue(object obj)
+		{
+			return prop.GetValue(obj, null);
+		}
+
+		internal override void SetValue(object obj, object val)
+		{
+			if(FieldTypeWrapper.IsGhost)
+			{
+				object temp = GetValue(obj);
+				FieldTypeWrapper.GhostRefField.SetValue(temp, val);
+				val = temp;
+			}
+			prop.SetValue(obj, val, null);
+		}
+#endif
+
 #if !COMPACT_FRAMEWORK
 		protected override void EmitGetImpl(ILGenerator ilgen)
 		{
@@ -1541,7 +1600,7 @@ namespace IKVM.Internal
 			// NOTE even though you're not supposed to access a constant static final (the compiler is supposed
 			// to inline them), we have to support it (because it does happen, e.g. if the field becomes final
 			// after the referencing class was compiled, or when we're accessing an unsigned primitive .NET field)
-			object v = GetValue(null);
+			object v = GetConstantValue();
 			if(v == null)
 			{
 				ilgen.Emit(OpCodes.Ldnull);
@@ -1605,15 +1664,39 @@ namespace IKVM.Internal
 			ilgen.Emit(OpCodes.Pop);
 		}
 #endif
-
-		internal override object GetValue(object obj)
+		internal object GetConstantValue()
 		{
 			if(constant == null)
 			{
+#if WHIDBEY
+				FieldInfo field = GetField();
+#if !STATIC_COMPILER
+				if(field.FieldType.IsEnum && !field.DeclaringType.IsEnum)
+				{
+					if(field.DeclaringType.Assembly.ReflectionOnly)
+					{
+						return null;
+					}
+					constant = field.GetValue(null);
+				}
+				else
+#endif // !STATIC_COMPILER
+				{
+					constant = field.GetRawConstantValue();
+				}
+#else // WHIDBEY
 				constant = GetField().GetValue(null);
+#endif // WHIDBEY
 			}
 			return constant;
 		}
+
+		internal override object GetValue(object obj)
+		{
+			// on a non-broken CLR GetValue on a literal will not trigger type initialization, but on Java it should
+			DeclaringType.RunClassInit();
+			return GetConstantValue();
+		}
 	}
 
 #if !COMPACT_FRAMEWORK
@@ -1633,13 +1716,13 @@ namespace IKVM.Internal
 		internal void DoLink(TypeBuilder typeBuilder)
 		{
 			basefield.Link();
-			if(basefield.IsLiteralField)
+			if(basefield is ConstantFieldWrapper)
 			{
 				FieldAttributes attribs = basefield.IsPublic ? FieldAttributes.Public : FieldAttributes.FamORAssem;
 				attribs |= FieldAttributes.Static | FieldAttributes.Literal;
 				FieldBuilder fb = typeBuilder.DefineField(Name, basefield.FieldTypeWrapper.TypeAsSignatureType, attribs);
 				AttributeHelper.HideFromReflection(fb);
-				fb.SetConstant(basefield.GetValue(null));
+				fb.SetConstant(((ConstantFieldWrapper)basefield).GetConstantValue());
 			}
 			else
 			{
@@ -1680,7 +1763,7 @@ namespace IKVM.Internal
 
 		protected override void EmitGetImpl(CountingILGenerator ilgen)
 		{
-			if(basefield.IsLiteralField)
+			if(basefield is ConstantFieldWrapper)
 			{
 				basefield.EmitGet(ilgen);
 			}
