@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002, 2003, 2004, 2005, 2006 Jeroen Frijters
+  Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -264,7 +264,7 @@ namespace IKVM.NativeCode.java
 						}
 						else
 						{
-							instanceTypeWrapper = IKVM.Runtime.Util.GetTypeWrapperFromObject(instance);
+							instanceTypeWrapper = IKVM.NativeCode.ikvm.runtime.Util.GetTypeWrapperFromObject(instance);
 						}
 						return member.IsAccessibleFrom(member.DeclaringType, (TypeWrapper)callerTypeWrapper, instanceTypeWrapper);
 					}
@@ -454,90 +454,10 @@ namespace IKVM.NativeCode.java
 			{
 				ByteCodeHelper.arraycopy(src, srcStart, dest, destStart, len);
 			}
-
-			public static void setErr(object printStream)
-			{
-				TypeWrapper tw = ClassLoaderWrapper.LoadClassCritical("java.lang.System");
-				FieldWrapper fw = tw.GetFieldWrapper("err", "Ljava.io.PrintStream;");
-				fw.SetValue(null, printStream);
-			}
-
-			public static void setIn(object inputStream)
-			{
-				TypeWrapper tw = ClassLoaderWrapper.LoadClassCritical("java.lang.System");
-				FieldWrapper fw = tw.GetFieldWrapper("in", "Ljava.io.InputStream;");
-				fw.SetValue(null, inputStream);
-			}
-
-			public static void setOut(object printStream)
-			{
-				TypeWrapper tw = ClassLoaderWrapper.LoadClassCritical("java.lang.System");
-				FieldWrapper fw = tw.GetFieldWrapper("out", "Ljava.io.PrintStream;");
-				fw.SetValue(null, printStream);
-			}
 		}
 
 		public class VMClassLoader
 		{
-			public static Assembly findResourceAssembly(string name)
-			{
-				name = JVM.MangleResourceName(name);
-#if COMPACT_FRAMEWORK
-				// TODO
-				try
-				{
-					JVM.CoreAssembly.GetManifestResourceStream(name).Close();
-					return JVM.CoreAssembly;
-				}
-				catch(FileNotFoundException)
-				{
-					return null;
-				}
-#else
-				foreach(Assembly asm in AppDomain.CurrentDomain.GetAssemblies())
-				{
-					if(!(asm is System.Reflection.Emit.AssemblyBuilder))
-					{
-						if(asm.GetManifestResourceInfo(name) != null)
-						{
-							return asm;
-						}
-					}
-				}
-				return null;
-#endif
-			}
-
-			public static Assembly[] findResourceAssemblies(string name)
-			{
-				name = JVM.MangleResourceName(name);
-#if COMPACT_FRAMEWORK
-				// TODO figure this out
-				try
-				{
-					JVM.CoreAssembly.GetManifestResourceStream(name).Close();
-					return new Assembly[] { JVM.CoreAssembly };
-				}
-				catch(FileNotFoundException)
-				{
-					return new Assembly[0];
-				}
-#else
-				ArrayList list = new ArrayList();
-				foreach(Assembly asm in AppDomain.CurrentDomain.GetAssemblies())
-				{
-					if(!(asm is System.Reflection.Emit.AssemblyBuilder))
-					{
-						if(asm.GetManifestResourceInfo(name) != null)
-						{
-							list.Add(asm);
-						}
-					}
-				}
-				return (Assembly[])list.ToArray(typeof(Assembly));
-#endif
-			}
-
 			public static object loadClass(string name, bool resolve)
 			{
 				try
@@ -555,11 +475,6 @@ namespace IKVM.NativeCode.java
 				}
 			}
 
-			public static object getBootstrapClassLoader()
-			{
-				return ClassLoaderWrapper.GetJavaBootstrapClassLoader();
-			}
-
 			public static object getPrimitiveClass(char type)
 			{
 				switch(type)
@@ -589,20 +504,23 @@ namespace IKVM.NativeCode.java
 
 			public static object defineClassImpl(object classLoader, string name, byte[] data, int offset, int length, object protectionDomain)
 			{
-#if COMPACT_FRAMEWORK
-				throw JavaException.NoClassDefFoundError("Class loading is not supported on the Compact Framework");
-#else
 				Profiler.Enter("ClassLoader.defineClass");
 				try
 				{
 					try
 					{
-						ClassFile classFile = new ClassFile(data, offset, length, name);
+						ClassLoaderWrapper classLoaderWrapper = ClassLoaderWrapper.GetClassLoaderWrapper(classLoader);
+						ClassFileParseOptions cfp = ClassFileParseOptions.LineNumberTable;
+						if(classLoaderWrapper.EmitDebugInfo)
+						{
+							cfp |= ClassFileParseOptions.LocalVariableTable;
+						}
+						ClassFile classFile = new ClassFile(data, offset, length, name, cfp);
 						if(name != null && classFile.Name != name)
 						{
 							throw new NoClassDefFoundError(name + " (wrong name: " + classFile.Name + ")");
 						}
-						TypeWrapper type = ClassLoaderWrapper.GetClassLoaderWrapper(classLoader).DefineClass(classFile, protectionDomain);
+						TypeWrapper type = classLoaderWrapper.DefineClass(classFile, protectionDomain);
 						return type.ClassObject;
 					}
 					catch(RetargetableJavaException x)
@@ -614,33 +532,6 @@ namespace IKVM.NativeCode.java
 				{
 					Profiler.Leave("ClassLoader.defineClass");
 				}
-#endif
-			}
-
-			public static string getPackageName(Type type)
-			{
-				string name;
-#if !COMPACT_FRAMEWORK
-				if(type.Assembly is System.Reflection.Emit.AssemblyBuilder)
-				{
-					name = ClassLoaderWrapper.GetWrapperFromType(type).Name;
-				}
-				else
-#endif
-					if(type.Module.IsDefined(typeof(JavaModuleAttribute), false))
-				{
-					name = CompiledTypeWrapper.GetName(type);
-				}
-				else
-				{
-					name = DotNetTypeWrapper.GetName(type);
-				}
-				int dot = name.LastIndexOf('.');
-				if(dot > 0)
-				{
-					return name.Substring(0, dot);
-				}
-				return null;
 			}
 
 			public static object findLoadedClass(object javaClassLoader, string name)
@@ -651,25 +542,13 @@ namespace IKVM.NativeCode.java
 				{
 					return wrapper.ClassObject;
 				}
-				// For compatibility with Sun's JDK 1.4, we also find array variations of already loaded classes
-				if(name.StartsWith("[") && name.EndsWith(";"))
-				{
-					int rank = 1;
-					while(name[rank] == '[')
-					{
-						rank++;
-					}
-					if(name[rank] == 'L' && name[rank + 1] != '[')
-					{
-						wrapper = loader.GetLoadedClass(name.Substring(rank + 1, name.Length - (rank + 2)));
-						if(wrapper != null)
-						{
-							return wrapper.MakeArrayType(rank).ClassObject;
-						}
-					}
-				}
 				return null;
 			}
+
+			public static object getAssemblyClassLoader(Assembly asm)
+			{
+				return ClassLoaderWrapper.GetAssemblyClassLoader(asm).GetJavaClassLoader();
+			}
 		}
 
 		public class VMClass
@@ -735,7 +614,7 @@ namespace IKVM.NativeCode.java
 
 			public static object getClassFromType(Type type)
 			{
-				return IKVM.Runtime.Util.GetClassFromTypeHandle(type.TypeHandle);
+				return IKVM.NativeCode.ikvm.runtime.Util.getClassFromTypeHandle(type.TypeHandle);
 			}
 
 			public static string GetName(object wrapper)
@@ -790,14 +669,7 @@ namespace IKVM.NativeCode.java
 			public static object getClassLoader0(object wrapper)
 			{
 				TypeWrapper tw = (TypeWrapper)wrapper;
-				object loader = tw.GetClassLoader().GetJavaClassLoader();
-				// HACK we have to exclude DynamicTypeWrapper instances, because proxies that are created by the bootstrap
-				// class loader also need to return null (but they don't live in the CoreAssembly)
-				if(loader == null && tw.Assembly != JVM.CoreAssembly && !(tw is DynamicTypeWrapper))
-				{
-					return ClassLoaderWrapper.GetSystemClassLoader().GetJavaClassLoader();
-				}
-				return loader;
+				return tw.GetClassLoader().GetJavaClassLoader();
 			}
 
 			public static object[] GetDeclaredMethods(object cwrapper, bool getMethods, bool publicOnly)
@@ -807,6 +679,16 @@ namespace IKVM.NativeCode.java
 				{
 					TypeWrapper wrapper = (TypeWrapper)cwrapper;
 					wrapper.Finish();
+					if(wrapper.HasVerifyError)
+					{
+						// TODO we should get the message from somewhere
+						throw new VerifyError();
+					}
+					if(wrapper.HasClassFormatError)
+					{
+						// TODO we should get the message from somewhere
+						throw new ClassFormatError(wrapper.Name);
+					}
 					// we need to look through the array for unloadable types, because we may not let them
 					// escape into the 'wild'
 					MethodWrapper[] methods = wrapper.GetMethods();
@@ -925,6 +807,10 @@ namespace IKVM.NativeCode.java
 						{
 							throw JavaException.NoClassDefFoundError(wrappers[i].Name);
 						}
+						if(!wrappers[i].IsAccessibleFrom(wrapper))
+						{
+							throw JavaException.IllegalAccessError("tried to access class {0} from class {1}", wrappers[i].Name, wrapper.Name);
+						}
 						innerclasses[i] = wrappers[i].ClassObject;
 					}
 					return innerclasses;
@@ -951,7 +837,19 @@ namespace IKVM.NativeCode.java
 					{
 						throw JavaException.NoClassDefFoundError(declaring.Name);
 					}
-					return declaring.ClassObject;
+					if(!declaring.IsAccessibleFrom(wrapper))
+					{
+						throw JavaException.IllegalAccessError("tried to access class {0} from class {1}", declaring.Name, wrapper.Name);
+					}
+					declaring.Finish();
+					foreach(TypeWrapper tw in declaring.InnerClasses)
+					{
+						if(tw == wrapper)
+						{
+							return declaring.ClassObject;
+						}
+					}
+					throw JavaException.IncompatibleClassChangeError("{0} and {1} disagree on InnerClasses attribute", declaring.Name, wrapper.Name);
 				}
 				catch(RetargetableJavaException x)
 				{
@@ -983,12 +881,15 @@ namespace IKVM.NativeCode.java
 				try
 				{
 					TypeWrapper wrapper = (TypeWrapper)cwrapper;
-					// NOTE unless ignoreInnerClassesAttribute is true, we don't return the modifiers from
-					// the TypeWrapper, because for inner classes the reflected modifiers are different
-					// from the physical ones
-					Modifiers modifiers = ignoreInnerClassesAttribute ?
-						wrapper.Modifiers : wrapper.ReflectiveModifiers;
-					return (int)modifiers;
+					if(ignoreInnerClassesAttribute)
+					{
+						// NOTE GNU Classpath's Class gets it wrong and sets the ignoreInnerClassesAttribute
+						// when it wants to find out about Enum, Annotation, etc. so we treat the flag instead
+						// to mean that we should return the real accessibility flags, but otherwise return the
+						// flags from the InnerClass attribute.
+						return (int)((wrapper.ReflectiveModifiers & ~Modifiers.AccessMask) | (wrapper.Modifiers & Modifiers.AccessMask));
+					}
+					return (int)wrapper.ReflectiveModifiers;
 				}
 				catch(RetargetableJavaException x)
 				{
@@ -1018,11 +919,6 @@ namespace IKVM.NativeCode.java
 						}
 						return enclosingClass.ClassObject;
 					}
-					if (wrapper.DeclaringTypeWrapper != null)
-					{
-						wrapper.DeclaringTypeWrapper.EnsureLoadable(wrapper.GetClassLoader());
-						return wrapper.DeclaringTypeWrapper.ClassObject;
-					}
 					return null;
 				}
 				catch(RetargetableJavaException x)
@@ -1109,10 +1005,10 @@ namespace IKVM.NativeCode.java
 				{
 					x.ToJava();
 				}
-				wrapper.RunClassInit();
 				Type type = wrapper.TypeAsTBD;
 				if(!type.IsArray && type.TypeInitializer != null)
 				{
+					wrapper.RunClassInit();
 					return !AttributeHelper.IsHideFromJava(type.TypeInitializer);
 				}
 				return false;
@@ -1216,6 +1112,19 @@ namespace IKVM.NativeCode.java
 #endif
 }
 
+namespace IKVM.NativeCode.gnu.java.lang.management
+{
+	public class VMClassLoadingMXBeanImpl
+	{
+		public static int getLoadedClassCount()
+		{
+			// we don't really have a number of classes loaded, but we'll
+			// return something anyway
+			return ClassLoaderWrapper.GetLoadedClassCount();
+		}
+	}
+}
+
 namespace IKVM.NativeCode.gnu.java.net.protocol.ikvmres
 {
 #if !WHIDBEY
@@ -1274,7 +1183,8 @@ namespace IKVM.NativeCode.gnu.java.net.protocol.ikvmres
 					stack[sp++] = char_tbl[cc - 256];
 					cc = ptr_tbl[cc - 256];
 				}
-				fc = stack[sp++] = cc;
+				stack[sp++] = cc;
+				fc = cc;
 				if (count >= 0) 
 				{
 					ptr_tbl[count] = prev;
@@ -1421,26 +1331,39 @@ namespace IKVM.NativeCode.gnu.java.net.protocol.ikvmres
 	{
 		public static Stream ReadResourceFromAssemblyImpl(Assembly asm, string resource)
 		{
+			// chop off the leading slash
+			resource = resource.Substring(1);
+			string mangledName = JVM.MangleResourceName(resource);
+			ManifestResourceInfo info = asm.GetManifestResourceInfo(mangledName);
+			if(info != null && info.FileName != null)
+			{
+				return asm.GetManifestResourceStream(mangledName);
+			}
 #if WHIDBEY
-			Stream s = asm.GetManifestResourceStream(JVM.MangleResourceName(resource.Substring(1)));
+			Stream s = asm.GetManifestResourceStream(mangledName);
 			if(s == null)
 			{
+				Tracer.Warning(Tracer.ClassLoading, "Resource \"{0}\" not found in {1}", resource, asm.FullName);
 				throw new FileNotFoundException("resource " + resource + " not found in assembly " + asm.FullName);
 			}
 			switch (s.ReadByte())
 			{
 				case 0:
+					Tracer.Info(Tracer.ClassLoading, "Reading resource \"{0}\" from {1}", resource, asm.FullName);
 					return s;
 				case 1:
+					Tracer.Info(Tracer.ClassLoading, "Reading compressed resource \"{0}\" from {1}", resource, asm.FullName);
 					return new System.IO.Compression.DeflateStream(s, System.IO.Compression.CompressionMode.Decompress, false);
 				default:
+					Tracer.Error(Tracer.ClassLoading, "Resource \"{0}\" in {1} has an unsupported encoding", resource, asm.FullName);
 					throw new IOException("Unsupported resource encoding for resource " + resource + " found in assembly " + asm.FullName);
 			}
 #else
-			using(Stream s = asm.GetManifestResourceStream(JVM.MangleResourceName(resource.Substring(1))))
+			using(Stream s = asm.GetManifestResourceStream(mangledName))
 			{
 				if(s == null)
 				{
+					Tracer.Warning(Tracer.ClassLoading, "Resource \"{0}\" not found in {1}", resource, asm.FullName);
 					throw new FileNotFoundException("resource " + resource + " not found in assembly " + asm.FullName);
 				}
 				using(System.Resources.ResourceReader r = new System.Resources.ResourceReader(s))
@@ -1449,22 +1372,52 @@ namespace IKVM.NativeCode.gnu.java.net.protocol.ikvmres
 					{
 						if((string)de.Key == "lz")
 						{
+							Tracer.Info(Tracer.ClassLoading, "Reading compressed resource \"{0}\" from {1}", resource, asm.FullName);
 							return new LZInputStream(new MemoryStream((byte[])de.Value));
 						}
 						else if((string)de.Key == "ikvm")
 						{
+							Tracer.Info(Tracer.ClassLoading, "Reading resource \"{0}\" from {1}", resource, asm.FullName);
 							return new MemoryStream((byte[])de.Value);
 						}
 						else
 						{
+							Tracer.Error(Tracer.ClassLoading, "Resource \"{0}\" in {1} has an unsupported encoding", resource, asm.FullName);
 							throw new IOException("Unsupported resource encoding " + de.Key + " for resource " + resource + " found in assembly " + asm.FullName);
 						}
 					}
+					Tracer.Error(Tracer.ClassLoading, "Resource \"{0}\" in {1} is invalid", resource, asm.FullName);
 					throw new IOException("Invalid resource " + resource + " found in assembly " + asm.FullName);
 				}
 			}
 #endif
 		}
+
+		public static object LoadClassFromAssembly(Assembly asm, string className)
+		{
+			TypeWrapper tw = ClassLoaderWrapper.GetAssemblyClassLoader(asm).LoadClassByDottedNameFast(className);
+			if(tw != null)
+			{
+				return tw.ClassObject;
+			}
+			return null;
+		}
+
+		public static Assembly LoadAssembly(string name)
+		{
+#if WHIDBEY
+			if(name.EndsWith("[ReflectionOnly]"))
+			{
+				return Assembly.ReflectionOnlyLoad(name.Substring(0, name.Length - 16));
+			}
+#endif
+			return Assembly.Load(name);
+		}
+
+		public static object GetGenericClassLoaderById(int id)
+		{
+			return ClassLoaderWrapper.GetGenericClassLoaderById(id).GetJavaClassLoader();
+		}
 	}
 }
 
@@ -1487,6 +1440,8 @@ namespace IKVM.NativeCode.gnu.classpath
 
 	public class VMStackWalker
 	{
+		private static readonly Hashtable isHideFromJavaCache = Hashtable.Synchronized(new Hashtable());
+
 		public static object getClassFromType(Type type)
 		{
 			TypeWrapper.AssertFinished(type);
@@ -1507,96 +1462,285 @@ namespace IKVM.NativeCode.gnu.classpath
 			// global methods have no type
 			if(type == null)
 			{
-				return JVM.Library.getSystemClassLoader();
+				return null;
 			}
 			else if(type.Module is System.Reflection.Emit.ModuleBuilder)
 			{
 				return ClassLoaderWrapper.GetWrapperFromType(type).GetClassLoader().GetJavaClassLoader();
 			}
-			else if(ClassLoaderWrapper.IsCoreAssemblyType(type))
-			{
-				return null;
-			}
 			else
 			{
-				return JVM.Library.getSystemClassLoader();
+				return ClassLoaderWrapper.GetAssemblyClassLoader(type.Assembly).GetJavaClassLoader();
 			}
 		}
 
 		public static Type getJNIEnvType()
 		{
+#if COMPACT_FRAMEWORK
+			return null;
+#else
 			return typeof(IKVM.Runtime.JNIEnv);
+#endif
 		}
 
 		public static bool isHideFromJava(MethodBase mb)
 		{
-			return mb.IsDefined(typeof(HideFromJavaAttribute), false)
-				|| mb.IsDefined(typeof(HideFromReflectionAttribute), false);
+			// TODO on .NET 2.0 isHideFromJavaCache should be a Dictionary<RuntimeMethodHandle, bool>
+			object cached = isHideFromJavaCache[mb];
+			if(cached == null)
+			{
+				cached = mb.IsDefined(typeof(HideFromJavaAttribute), false)
+					|| mb.IsDefined(typeof(HideFromReflectionAttribute), false);
+				isHideFromJavaCache[mb] = cached;
+			}
+			return (bool)cached;
 		}
 	}
 }
 
-namespace gnu.classpath
+namespace IKVM.NativeCode.ikvm.@internal
 {
-	// This type lives here, because we don't want unverifiable code in IKVM.GNU.Classpath
-	// (as that would prevents us from verifying it during the build process).
-#if !COMPACT_FRAMEWORK
-	[SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
-#endif
-	public unsafe sealed class Pointer
+	public class AssemblyClassLoader
 	{
-		// NOTE during static compilation this type is represented by a CompiledTypeWrapper,
-		// so we need to hide this field, because Java types cannot have pointer types.
-		// Note that at runtime this type is represented by a DotNetTypeWrapper and that
-		// DotNetTypeWrapper.LazyPublishMembers has a hack to prevent it from exposing any
-		// of the members of this type (because that would be a security problem).
-		[HideFromJava]
-		private byte* pb;
-
-		public Pointer(IntPtr p)
+		public static object LoadClass(object classLoader, string name)
 		{
-			this.pb = (byte*)p;
-		}
-
-		public IntPtr p()
-		{
-			return new IntPtr(pb);
-		}
-
-		public byte ReadByte(int index)
-		{
-			return pb[index];
-		}
-
-		public void WriteByte(int index, byte b)
-		{
-			pb[index] = b;
-		}
-
-		public void MoveMemory(int dst_offset, int src_offset, int count)
-		{
-			if(dst_offset < src_offset)
+			try
 			{
-				while(count-- > 0)
-					pb[dst_offset++] = pb[src_offset++];
+				ClassLoaderWrapper wrapper = classLoader == null ? ClassLoaderWrapper.GetBootstrapClassLoader() : (ClassLoaderWrapper)JVM.Library.getWrapperFromClassLoader(classLoader);
+				TypeWrapper tw = wrapper.LoadClassByDottedName(name);
+				Tracer.Info(Tracer.ClassLoading, "Loaded class \"{0}\" from {1}", name, classLoader == null ? "boot class loader" : classLoader);
+				return tw.ClassObject;
 			}
-			else
+			catch(RetargetableJavaException x)
 			{
-				dst_offset += count;
-				src_offset += count;
-				while(count-- > 0)
-					pb[--dst_offset] = pb[--src_offset];
+				Tracer.Info(Tracer.ClassLoading, "Failed to load class \"{0}\" from {1}", name, classLoader == null ? "boot class loader" : classLoader);
+				throw x.ToJava();
+			}
+		}
+
+		public static Assembly[] FindResourceAssemblies(object classLoader, string name, bool firstOnly)
+		{
+			IKVM.Internal.AssemblyClassLoader wrapper = classLoader == null ? ClassLoaderWrapper.GetBootstrapClassLoader() : (JVM.Library.getWrapperFromClassLoader(classLoader) as IKVM.Internal.AssemblyClassLoader);
+			if(wrapper == null)
+			{
+				// must be a GenericClassLoader
+				Tracer.Info(Tracer.ClassLoading, "Failed to find resource \"{0}\" in generic class loader", name);
+				return null;
+			}
+			Assembly[] assemblies = wrapper.FindResourceAssemblies(name, firstOnly);
+			if(assemblies == null || assemblies.Length == 0)
+			{
+				Tracer.Info(Tracer.ClassLoading, "Failed to find resource \"{0}\" in {1}", name, wrapper.Assembly.FullName);
+				return null;
+			}
+			foreach(Assembly asm in assemblies)
+			{
+				Tracer.Info(Tracer.ClassLoading, "Found resource \"{0}\" in {1}", name, asm.FullName);
+			}
+			return assemblies;
+		}
+
+		// NOTE the array may contain duplicates!
+		public static string[] GetPackages(object classLoader)
+		{
+			IKVM.Internal.AssemblyClassLoader wrapper = classLoader == null ? ClassLoaderWrapper.GetBootstrapClassLoader() : (JVM.Library.getWrapperFromClassLoader(classLoader) as IKVM.Internal.AssemblyClassLoader);
+			if(wrapper == null)
+			{
+				// must be a GenericClassLoader
+				return null;
+			}
+			string[] packages = new string[0];
+			foreach(Module m in wrapper.Assembly.GetModules(false))
+			{
+				object[] attr = m.GetCustomAttributes(typeof(PackageListAttribute), false);
+				foreach(PackageListAttribute p in attr)
+				{
+					string[] mp = p.GetPackages();
+					string[] tmp = new string[packages.Length + mp.Length];
+					Array.Copy(packages, 0, tmp, 0, packages.Length);
+					Array.Copy(mp, 0, tmp, packages.Length, mp.Length);
+					packages = tmp;
+				}
+			}
+			return packages;
+		}
+
+		public static bool IsReflectionOnly(Assembly asm)
+		{
+#if WHIDBEY
+			return asm.ReflectionOnly;
+#else
+			return false;
+#endif
+		}
+
+		public static int GetGenericClassLoaderId(object classLoader)
+		{
+			return ClassLoaderWrapper.GetGenericClassLoaderId((ClassLoaderWrapper)JVM.Library.getWrapperFromClassLoader(classLoader));
+		}
+
+		public static Assembly GetBootClassLoaderAssembly()
+		{
+			return ((IKVM.Internal.AssemblyClassLoader)ClassLoaderWrapper.GetBootstrapClassLoader()).Assembly;
+		}
+
+		public static string GetGenericClassLoaderName(object classLoader)
+		{
+			return ((GenericClassLoader)JVM.Library.getWrapperFromClassLoader(classLoader)).GetName();
+		}
+	}
+
+	namespace stubgen
+	{
+		public class StubGenerator
+		{
+			public static string getAssemblyName(object c)
+			{
+				ClassLoaderWrapper loader = TypeWrapper.FromClass(c).GetClassLoader();
+				IKVM.Internal.AssemblyClassLoader acl = loader as IKVM.Internal.AssemblyClassLoader;
+				if(acl != null)
+				{
+					return acl.Assembly.FullName;
+				}
+				else
+				{
+					return ((IKVM.Internal.GenericClassLoader)loader).GetName();
+				}
+			}
+
+			public static object getFieldConstantValue(object fieldWrapper)
+			{
+				return ((FieldWrapper)fieldWrapper).GetConstant();
+			}
+
+			public static bool isFieldDeprecated(object fieldWrapper)
+			{
+				FieldInfo fi = ((FieldWrapper)fieldWrapper).GetField();
+				if(fi != null)
+				{
+					return AttributeHelper.IsDefined(fi, typeof(ObsoleteAttribute));
+				}
+				GetterFieldWrapper getter = fieldWrapper as GetterFieldWrapper;
+				if(getter != null)
+				{
+					return AttributeHelper.IsDefined(getter.GetProperty(), typeof(ObsoleteAttribute));
+				}
+				return false;
+			}
+
+			public static bool isMethodDeprecated(object methodWrapper)
+			{
+				MethodBase mb = ((MethodWrapper)methodWrapper).GetMethod();
+				return mb != null && AttributeHelper.IsDefined(mb, typeof(ObsoleteAttribute));
+			}
+
+			public static bool isClassDeprecated(object wrapper)
+			{
+				Type type = ((TypeWrapper)wrapper).TypeAsTBD;
+				// we need to check type for null, because ReflectionOnly
+				// generated delegate inner interfaces don't really exist
+				return type != null && AttributeHelper.IsDefined(type, typeof(ObsoleteAttribute));
 			}
 		}
 	}
 }
 
+namespace IKVM.NativeCode.ikvm.runtime
+{
+	public class Util
+	{
+		private static Type enumEnumType = JVM.CoreAssembly.GetType("ikvm.internal.EnumEnum");
+		private static FieldInfo enumEnumTypeField = enumEnumType.GetField("typeWrapper", BindingFlags.Instance | BindingFlags.NonPublic);
+
+		// we don't want "beforefieldinit"
+		static Util() {}
+
+		public static object getClassFromObject(object o)
+		{
+			return GetTypeWrapperFromObject(o).ClassObject;
+		}
+
+		internal static TypeWrapper GetTypeWrapperFromObject(object o)
+		{
+			Type t = o.GetType();
+			if(t.IsPrimitive || (ClassLoaderWrapper.IsRemappedType(t) && !t.IsSealed))
+			{
+				return DotNetTypeWrapper.GetWrapperFromDotNetType(t);
+			}
+			if(t == enumEnumType)
+			{
+				return (TypeWrapper)enumEnumTypeField.GetValue(o);
+			}
+			return ClassLoaderWrapper.GetWrapperFromType(t);
+		}
+
+		public static object getClassFromTypeHandle(RuntimeTypeHandle handle)
+		{
+			Type t = Type.GetTypeFromHandle(handle);
+			if(t.IsPrimitive || ClassLoaderWrapper.IsRemappedType(t) || t == typeof(void))
+			{
+				return DotNetTypeWrapper.GetWrapperFromDotNetType(t).ClassObject;
+			}
+			if(Whidbey.ContainsGenericParameters(t))
+			{
+				return null;
+			}
+			TypeWrapper tw = ClassLoaderWrapper.GetWrapperFromType(t);
+			if(tw != null)
+			{
+				return tw.ClassObject;
+			}
+			return null;
+		}
+
+		public static object getFriendlyClassFromType(Type type)
+		{
+			if(Whidbey.ContainsGenericParameters(type))
+			{
+				return null;
+			}
+			int rank = 0;
+			while(type.IsArray)
+			{
+				type = type.GetElementType();
+				rank++;
+			}
+			if(type.DeclaringType != null
+				&& AttributeHelper.IsGhostInterface(type.DeclaringType))
+			{
+				type = type.DeclaringType;
+			}
+			TypeWrapper wrapper = ClassLoaderWrapper.GetWrapperFromType(type);
+			if(rank > 0)
+			{
+				wrapper = wrapper.MakeArrayType(rank);
+			}
+			return wrapper.ClassObject;
+		}
+
+		public static Type GetInstanceTypeFromTypeWrapper(object wrapperObject)
+		{
+			TypeWrapper wrapper = (TypeWrapper)wrapperObject;
+			if(wrapper.IsDynamicOnly)
+			{
+				return null;
+			}
+			if(wrapper.IsRemapped && wrapper.IsFinal)
+			{
+				return wrapper.TypeAsTBD;
+			}
+			return wrapper.TypeAsBaseType;
+		}
+	}
+}
+
+#if FIRST_PASS
 namespace ikvm.@internal
 {
 	public interface LibraryVMInterface
 	{
 		object loadClass(object classLoader, string name);
-		object newClass(object wrapper, object protectionDomain);
+		object newClass(object wrapper, object protectionDomain, object classLoader);
 		object getWrapperFromClass(object clazz);
 		object getWrapperFromField(object field);
 		object getWrapperFromMethodOrConstructor(object methodOrConstructor);
@@ -1604,8 +1748,6 @@ namespace ikvm.@internal
 		object getWrapperFromClassLoader(object classLoader);
 		void setWrapperForClassLoader(object classLoader, object wrapper);
 
-		object getSystemClassLoader();
-
 		object box(object val);
 		object unbox(object val);
 
@@ -1659,5 +1801,8 @@ namespace ikvm.@internal
 		Exception newInstantiationException(string msg);
 		Exception newInterruptedException();
 		Exception newIllegalMonitorStateException();
+
+		object newAssemblyClassLoader(Assembly asm);
 	}
 }
+#endif // !FIRST_PASS
